/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.animal;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.List;

import org.apache.uima.jcas.JCas;

import de.unibi.techfak.scie.annotators.quantities.Unit;
import de.unibi.techfak.scie.annotators.quantities.UnitQuantityAnnotator;
import de.unibi.techfak.isy.lppba.descriptors.Quantity;
import de.unibi.techfak.isy.lppba.descriptors.Weight;
import de.unibi.techfak.isy.lppba.descriptors.Token;

/**
 * This annotates weights.
 * 
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class WeightAnnotator extends UnitQuantityAnnotator {

	private static final Unit[] units = { Unit.MILLIGRAMMES, Unit.GRAMMES,
			Unit.KILOGRAMMES, Unit.TONNES, Unit.POUNDS };

	public static double convertToSIValue(final Unit currentUnit,
			final double currentValue) {
		switch (currentUnit) {
		case MILLIGRAMMES:
			return convertToSIValue(Unit.GRAMMES, currentValue / 1000);
		case GRAMMES:
			return currentValue / 1000;
		case KILOGRAMMES:
			return currentValue;
		case TONNES:
			return currentValue * 1000;
		case POUNDS:
			return currentValue * 0.453592;
		default:
			throw new RuntimeException("Unrecognized Weight Unit: " + currentUnit);
		}
	}

	public WeightAnnotator() {
		super(units);
	}

	@Override
	public void handleQuantity(Quantity q, Unit unit, List<Token> unitTokens,
			JCas jcas) {
		Token unitToken = unitTokens.get(0);
		// calculate the standard value if possible
		final double SIValue;
		if (q instanceof de.unibi.techfak.isy.lppba.descriptors.Integer) {
			SIValue = convertToSIValue(unit,
					((de.unibi.techfak.isy.lppba.descriptors.Integer) q).getValue());
			if (SIValue < 0) {
				return;
			}
		} else if (q instanceof de.unibi.techfak.isy.lppba.descriptors.Double) {
			SIValue = convertToSIValue(unit,
					((de.unibi.techfak.isy.lppba.descriptors.Double) q).getValue());
			if (SIValue < 0) {
				return;
			}
		} else {
			SIValue = -1;
		}
		// then create the annotation.
		final int begin = q.getBegin();
		final int end = unitToken.getEnd();
		final Weight weight = new Weight(jcas, begin, end);
		weight.setUnitName(unit.getNormalizedName());
		weight.setValue(q);
		weight.setSIValue(SIValue);
		weight.setAnnotationId(AnnotationCounter.getUniqueId());
		weight.addToIndexes();
	};
}
