/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.injury;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.isy.lppba.descriptors.Duration;
import de.unibi.techfak.isy.lppba.descriptors.Injury;
import de.unibi.techfak.isy.lppba.descriptors.InjuryDevice;
import de.unibi.techfak.isy.lppba.descriptors.InjuryHeight;
import de.unibi.techfak.isy.lppba.descriptors.InjuryType;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMACompositor;
import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;
import java.util.ArrayList;
import java.util.Collection;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

public class InjuryAnnotator extends org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public InjuryAnnotator() {
	}
	public static final InjuryHeightModel HEIGHTMODEL = new InjuryHeightModel();
	public static final InjuryDeviceModel DEVICEMODEL = new InjuryDeviceModel();
	//public static final InjuryDeviceModel INJURYDEVICEMODEL = new InjuryDeviceModel();
	public static final DurationModel DURATIONMODEL = new DurationModel();
	public static final UIMASemanticProbabilityModel[] MODELS = {HEIGHTMODEL,
		DEVICEMODEL, DURATIONMODEL};
	public static final Class[] CLASSES = {InjuryHeight.class, InjuryDevice.class,
		Duration.class};
	public static final InjuryCompositor COMPOSITOR = new InjuryCompositor();
	/**
	 * This specifies about how far we search for annotations that belong to a Injury. For more
	 * details on this topic have a look at the AnnotationCompositionUtil class.
	 */
	public static final double SIGMA = 15;
	/**
	 * Injury annotations below this probability threshold are thrown away. This threshold was
	 * derived by using statistics over the available 20 test papers.
	 */
	public static final double THRESHOLD = 0.2;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final UIMAWordSyntacticProbabilityModel syntacticModel
			= new UIMAWordSyntacticProbabilityModel(SIGMA, jcas);
		COMPOSITOR.setSyntacticModel(syntacticModel);
		final Collection<InjuryType> types = JCasUtil.select(jcas, InjuryType.class);
		for (final InjuryType type : types) {
			UIMAAnnotationCompositionUtil.calculateComposita(type, syntacticModel, MODELS, CLASSES,
				COMPOSITOR, jcas);
		}
	}

	/**
	 * The likelihood model for InjuryHeight.
	 */
	public static class InjuryHeightModel extends UIMASemanticProbabilityModel<InjuryHeight> {

		public InjuryHeightModel() {
		}

		@Override
		public double calculateSemanticProbability(InjuryHeight candidate) {
			return candidate.getProbability();
		}
	}

	/**
	 * The likelihood model for Durations.
	 */
	public static class DurationModel extends UIMASemanticProbabilityModel<Duration> {

		public DurationModel() {
		}

		@Override
		public double calculateSemanticProbability(Duration candidate) {
			//every Duration is equally probable.
			return 1;
		}
	}

	/**
	 * The likelihood model for InjuryDevice.
	 */
	public static class InjuryDeviceModel extends UIMASemanticProbabilityModel<InjuryDevice> {

		public InjuryDeviceModel() {
		}

		@Override
		public double calculateSemanticProbability(InjuryDevice candidate) {
			return candidate.getProbability();
		}
	}

	public static class InjuryCompositor extends UIMACompositor<Injury> {

		private SyntacticProbabilityModel syntacticModel;

		public InjuryCompositor() {
		}

		public void setSyntacticModel(SyntacticProbabilityModel syntacticModel) {
			this.syntacticModel = syntacticModel;
		}

		public SyntacticProbabilityModel getSyntacticModel() {
			return syntacticModel;
		}

		@Override
		public Injury UIMACompose(Annotation core,
			ArrayList<ProbableHypothesis> slotCandidates,
			int minBegin, int maxEnd, JCas jcas) {
			final ProbableHypothesis<UIMAAnnotationCandidate> heightHypo = slotCandidates.get(0);
			final ProbableHypothesis<UIMAAnnotationCandidate> deviceHypo = slotCandidates.get(1);
			final ProbableHypothesis<UIMAAnnotationCandidate> durHypo = slotCandidates.get(2);
			//calculate the overall probability for this annotation which we arbitrarily define as
			//the average probability of the annotations divided by a normalization factor.
			double probability = 0;
			for (final ProbableHypothesis<UIMAAnnotationCandidate> slotHypo
				: slotCandidates) {
				if (slotHypo != null) {
					probability += slotHypo.getProbability();
				}
			}
			probability = probability / slotCandidates.size();
			//and we cut off the probability at 1.
			if (probability > 1) {
				probability = 1;
			}
			//if the composition probability is bigger than our threshold, store
			//this result
			if (probability > THRESHOLD) {
				//create annotation
				final Injury inj = new Injury(jcas, minBegin, maxEnd);
				inj.setAnnotationId(AnnotationCounter.getUniqueId());
				inj.setInjuryType((InjuryType) core);
				//get all slot fillers
				if (heightHypo != null && heightHypo.getCandidate() != null) {
					inj.setInjuryHeight((InjuryHeight) heightHypo.getCandidate().getAnnotation());
				}
				if (deviceHypo != null && deviceHypo.getCandidate() != null) {
					inj.setInjuryDevice((InjuryDevice) deviceHypo.getCandidate().getAnnotation());
				}
				if (durHypo != null && durHypo.getCandidate() != null) {
					inj.setDuration((Duration) durHypo.getCandidate().getAnnotation());
				}
				inj.setProbability(probability);
				inj.addToIndexes();
				return inj;
			} else {
				return null;
			}
		}
	}
}
