/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.quantities;

import de.unibi.techfak.scie.annotators.AnnotationCounter;

import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.component.JCasAnnotator_ImplBase;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

import de.unibi.techfak.isy.lppba.descriptors.Number;
import de.unibi.techfak.isy.lppba.descriptors.Range;
import de.unibi.techfak.isy.lppba.descriptors.Token;

/**
 * This annotates ranges of numbers such as 4 - 5 or 30 to 40. This annotator
 * depends on the NumberAnnotator and WordNumberAnnotator.
 * 
 * @author Raphael Dickfelder - rdickfel@techfak.uni-bielefeld.de
 */
public class RangeAnnotator extends JCasAnnotator_ImplBase {

	/**
	 * This specifies which intermediate content is allowed between two numbers to
	 * regard it as a range. Unfortunately it is not possible to specify -
	 * directly as there are different kinds of - that can not be specified here
	 * in the code as it seems.
	 */
//	private static final Pattern RANGE = Pattern
//			.compile("(\\s*[^\\,\\.\\!\\w\\(\\)\\[\\]\\=\\s]\\s*|\\s+to\\s+)");
	private static final String STROKES=StringEscapeUtils.unescapeHtml("&#x2D;&#x2212;&#x2010;&#x2013;&#x2014;&#x2E3A;");
	private static final Pattern RANGE = Pattern
			.compile("(\\s*["+STROKES+"]\\s*|\\s+to\\s+)");
	
	private static final String BETWEEN= "between";
	private static final String AND= "and";

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final Collection<Number> numbers = JCasUtil.select(jcas, Number.class);
		for (Number n : numbers) {
			List<Number> followNumber = JCasUtil.selectFollowing(jcas, Number.class,
					n, 2);
			for (Number fol : followNumber) {
				final String between = jcas.getDocumentText().substring(n.getEnd(),
						fol.getBegin());
				final Matcher matcher = RANGE.matcher(between);
				if (matcher.matches() && n.getClass() == fol.getClass()) {
					final Range range = new Range(jcas);
					range.setBegin(n.getBegin());
					range.setEnd(fol.getEnd());
					range.setLowVal(n);
					range.setHighVal(fol);
					range.setAnnotationId(AnnotationCounter.getUniqueId());
					range.addToIndexes();
				}else{
					List<Token> before=JCasUtil.selectPreceding(jcas, Token.class,n, 1);
					if(!before.isEmpty()&& before.get(0).getCoveredText().trim().toLowerCase().equals(BETWEEN)){
						if(between.trim().toLowerCase().equals(AND)){
							final Range range = new Range(jcas);
							range.setBegin(before.get(0).getBegin());
							range.setEnd(fol.getEnd());
							range.setLowVal(n);
							range.setHighVal(fol);
							range.setAnnotationId(AnnotationCounter.getUniqueId());
							range.addToIndexes();
						}
					}
				}
				
			}

		}
	}
}
