/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.result;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.unibi.techfak.isy.lppba.descriptors.Result;
import de.unibi.techfak.isy.lppba.descriptors.InvestigationMethods;
import de.unibi.techfak.isy.lppba.descriptors.PValue;
import de.unibi.techfak.isy.lppba.descriptors.Significance;
import de.unibi.techfak.isy.lppba.descriptors.Trend;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMACompositor;
import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import java.util.HashSet;

public class ResultAnnotator extends org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public ResultAnnotator() {
	}
	public static final SignificanceModel SIGNIFICANCE = new SignificanceModel();
	public static final TrendModel TREND = new TrendModel();
	public static final PValueModel PVALUE = new PValueModel();
	public static final UIMASemanticProbabilityModel[] MODELS = {SIGNIFICANCE, TREND, PVALUE};
	public static final Class[] CLASSES = {Significance.class, Trend.class, PValue.class};
	public static final ResultCompositor COMPOSITOR = new ResultCompositor();
	/**
	 * This specifies about how far we search for annotations that belong to a OutCome. For more
	 * details on this topic have a look at the AnnotationCompositionUtil class.
	 */
	public static final double SIGMA = 15;
	/**
	 * OutCome annotations below this probability threshold are thrown away. This threshold was
	 * derived by using statistics over the available 20 test papers.
	 */
	public static final double THRESHOLD = 0.2;
	/**
	 * This determines how many aggregations in the same text we accept at max. If there are more
	 * the most probable ones will be chosen.
	 */
	public static final int MAXOVERLAPPING = 5;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final UIMAWordSyntacticProbabilityModel syntacticModel =
				new UIMAWordSyntacticProbabilityModel(SIGMA, jcas);
		COMPOSITOR.setSyntacticModel(syntacticModel);
		final Collection<InvestigationMethods> methods = JCasUtil.select(jcas,
				InvestigationMethods.class);
		for (final InvestigationMethods method : methods) {
			UIMAAnnotationCompositionUtil.calculateComposita(method, syntacticModel,
					MODELS, CLASSES, COMPOSITOR, jcas);
		}
	}

	/**
	 * The likelihood model for the Significance.
	 */
	public static class SignificanceModel extends UIMASemanticProbabilityModel<Significance> {

		public SignificanceModel() {
		}

		@Override
		public double calculateSemanticProbability(Significance candidate) {
			return candidate.getProbability();
		}
	}

	/**
	 * The likelihood model for the Trend.
	 */
	public static class TrendModel extends UIMASemanticProbabilityModel<Trend> {

		public TrendModel() {
		}

		@Override
		public double calculateSemanticProbability(Trend candidate) {
			return candidate.getProbability();
		}
	}

	/**
	 * The trivial likelihood Model for OutComePart.
	 */
	public static class PValueModel extends UIMASemanticProbabilityModel<PValue> {

		public PValueModel() {
		}

		@Override
		public double calculateSemanticProbability(PValue candidate) {
			// every PValue is equally probable.
			return 1;
		}
	}

	public static class ResultCompositor extends UIMACompositor<Result> {

		private SyntacticProbabilityModel syntacticModel;

		public ResultCompositor() {
		}

		public void setSyntacticModel(SyntacticProbabilityModel syntacticModel) {
			this.syntacticModel = syntacticModel;
		}

		public SyntacticProbabilityModel getSyntacticModel() {
			return syntacticModel;
		}

		@Override
		public Result UIMACompose(Annotation core,
				ArrayList<ProbableHypothesis> slotCandidates, int minBegin, int maxEnd,
				JCas jcas) {
			final ProbableHypothesis<UIMAAnnotationCandidate> sigHypo = slotCandidates
					.get(0);
			final ProbableHypothesis<UIMAAnnotationCandidate> trendHypo = slotCandidates
					.get(1);
			final ProbableHypothesis<UIMAAnnotationCandidate> pValueHypo = slotCandidates
					.get(2);
			// calculate the overall probability for this annotation which we
			// arbitrarily define as
			// the average probability of the annotations divided by a normalization
			// factor.
			double probability = 0;
			for (final ProbableHypothesis<UIMAAnnotationCandidate> slotHypo : slotCandidates) {
				if (slotHypo != null) {
					probability += slotHypo.getProbability();
				}
			}
			probability = probability / slotCandidates.size();
			// and we cut off the probability at 1.
			if (probability > 1) {
				probability = 1;
			}
			// if the composition probability is bigger than our threshold, store
			// this result
			if (probability > THRESHOLD) {
				/*
				 * Before we create an annotation we want to check whether there are other result
				 * annoations overlapping with this one already.
				 */
				final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
				final HashSet<Result> overlappings = new HashSet<>();
				overlappings.addAll(jcasUtil.selectCovering(Result.class, minBegin, minBegin + 1));
				overlappings.addAll(jcasUtil.selectCovering(Result.class, maxEnd - 1, maxEnd));
				if (overlappings.size() >= MAXOVERLAPPING) {
					//get the annotation with minimum probability.
					double minProb = 1;
					Result minAnno = null;
					for (final Result overlapping : overlappings) {
						if (overlapping.getProbability() < minProb) {
							minAnno = overlapping;
							minProb = overlapping.getProbability();
						}
					}
					if (minProb < probability) {
						/*
						 * If the existing minimum probability aggregation in that region has a lower
						 * probability than our new annotation, we kick the old one out.
						 */
						minAnno.removeFromIndexes();
						jcasUtil.remove(Result.class, minAnno);
					} else {
						//otherwise we do not add the new one.
						return null;
					}
				}

				// create annotation
				final Result out = new Result(jcas, minBegin, maxEnd);
				out.setInvestigationMethods((InvestigationMethods) core);
				// get all slot fillers
				if (sigHypo != null && sigHypo.getCandidate() != null) {
					out.setSignificance((Significance) sigHypo.getCandidate().getAnnotation());
				}
				if (trendHypo != null && trendHypo.getCandidate() != null) {
					out.setTrend((Trend) trendHypo.getCandidate().getAnnotation());
				}
				if (pValueHypo != null && pValueHypo.getCandidate() != null) {
					out.setPValue((PValue) pValueHypo.getCandidate().getAnnotation());
				}
				out.setProbability(probability);
				out.setAnnotationId(AnnotationCounter.getUniqueId());
				out.addToIndexes();
				jcasUtil.add(Result.class, out);
				return out;
			} else {
				return null;
			}
		}
	}
}