/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.db;

import de.unibi.techfak.scie.configuration.Configuration;
import de.unibi.techfak.scie.configuration.ConfigurationEntry;
import de.unibi.techfak.scie.configuration.ConfigurationSet;
import de.unibi.techfak.scie.configuration.ConfigurationSingleton;
import de.unibi.techfak.scie.configuration.ValidityException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

/**
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class DatabaseSingleton {

	private static final Configuration configuration = ConfigurationSingleton.getConfiguration();

	private static final Set<String> availableDatabases = new HashSet();

	private static DatabaseInstance instance = null;

	private static boolean initialized = false;

	private static void registerDB(String dbName, String jdbcClass,
			Collection<ConfigurationEntry> configSet) {
		try {
			// Check whether the given class exists (otherwise fails with an
			// exception)
			Class.forName(jdbcClass);

			// Register the configuration set and mark this db as "available"
			configuration.registerConfigurationSet(dbName, configSet);
			availableDatabases.add(dbName);
		} catch (ClassNotFoundException ex) {
		}
	}
	
	public static void initialize(String configFile) {
		// Load the (database) configuration
		initialize();

		try {
			ConfigurationSingleton.loadConfigurationIfExists(configFile);
		} catch (ValidityException ex) {
			throw new RuntimeException(ex);
		}	
	}


	public static void initialize() {
		// Abort if the singleton was already initialized
		if (initialized) {
			reset();
		}

		// Register the configuration of all database instances
		registerDB("postgresql", "org.postgresql.Driver",
				PostgreSQLDatabaseInstance.getConfigurationEntries());
		registerDB("mysql", "com.mysql.jdbc.Driver",
				MySQLDatabaseInstance.getConfigurationEntries());

		// Register the memory database
		availableDatabases.add("memory");

		// Insert the "db.type" configuration option
		configuration.registerConfigurationSet("db", new ConfigurationEntry[]{
			new ConfigurationEntry(
			"type",
			"The type of database that will be used",
			"memory",
			new ConfigurationEntry.ValidityChecker() {

				@Override
				public Boolean isValid(String value) throws ValidityException {
					return availableDatabases.contains(value);
				}
			})
		});

		// The singleton is now initialized
		initialized = true;
	}

	public static void reset() {
		if (initialized) {
			// Close the databse instance
			if (instance != null) {
				instance.close();
				instance = null;
			}

			// Make sure the configuration is reset
			configuration.reset();
		}
		initialized = false;
	}

	public static DatabaseInstance getInstance() {
		if (instance == null) {
			final String type = configuration.getValue("db", "type");
			final ConfigurationSet set = configuration.getSet(type);
			switch (type) {
				case "memory":
					instance = new MemoryDatabaseInstance(set);
					break;
				case "postgresql":
					instance = new PostgreSQLDatabaseInstance(set);
					break;
				case "mysql":
					instance = new MySQLDatabaseInstance(set);
					break;
				default:
					throw new RuntimeException("Cannot instanciate DB, unknown type " + type);
			}

			// Make sure the DB schema is up to date
			instance.ensureSchema();
		}

		return instance;
	}

}
