/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.util.LineSegment;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms.WordPosition;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * The Dictionary class is a helper class which is responsible for updating and
 * querying the ontology database dictionary.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
final class Dictionary {

	/**
	 * Threshold which contains the minimum match probability.
	 */
	private static final float MIN_MATCH_PROBABILITY = 0.5f;

	/**
	 * Reference to the database driver.
	 */
	private final OntologyDatabaseDriver driver;

	/**
	 * Constructor of the Dictionary class.
	 *
	 * @param driver reference to the database driver.
	 */
	public Dictionary(OntologyDatabaseDriver driver) {
		this.driver = driver;
	}

	/**
	 * Builds a mapping between the given words and their pseudo-phonetic keys.
	 *
	 * @param words is an arbitrary collection of input words.
	 * @return a mapping between the phonetic keys of the words and the
	 * corresponding words. Note that each word may have multiple phonetic keys
	 * attached to it and some words may not be in the resulting list at all as
	 * they were recognised as "stop word".
	 */
	static protected Map<String, Collection<String>> buildKeyWordMap(
			Collection<String> words) {
		// Create the resulting hash map and iterate over each input word
		Map<String, Collection<String>> res = new HashMap();
		for (String word : words) {
			// Calculate the pseudo phonetic keys of the stem and store the
			// relationship in the resulting map. Note that the keys have to be
			// built for the most special comparison mode, as it is impossible
			// to predict the comparison mode of the underlying node names.
			String[] keys = DictionaryAlgorithms.calculateWordKeys(word,
					ComparisonMode.STRICT);
			for (String key : keys) {
				if (!key.isEmpty()) {
					if (res.containsKey(key)) {
						res.get(key).add(word);
					} else {
						Collection<String> wordList = new ArrayList(1);
						wordList.add(word);
						res.put(key, wordList);
					}
				}
			}
		}
		return res;
	}

	/**
	 * Calculates the dictionary entries (mappings between the node handle and
	 * the corresponding pseudo-phonetic keys) for the given node. The mappings
	 * are filled into the given dictionary.
	 *
	 * @param descr is the node descriptor from which the corresponding words
	 * should be extracted.
	 * @param dict is the dictionary which will be filled in the process
	 * (entries are added to the given map).
	 */
	static protected void fillDictionaryEntriesForNode(NodeDescriptor descr,
			Map<NodeHandle, Set<String>> dict) {
		// Fill the dictionary with the names present in this node
		for (NodeName nameEntry : descr.getNames()) {
			final String name = nameEntry.getName();
			final Set<WordPosition> wordPositions = DictionaryAlgorithms.getWords(name);
			for (WordPosition wordPosition : wordPositions) {
				// Calculate the keys of the stem and store the relationship in
				// the result map.
				final String word = wordPosition.getWord();
				String[] ppks = DictionaryAlgorithms.calculateWordKeys(word,
						nameEntry.getComparisonMode());
				if (ppks.length > 0) {
					Set<String> keys = dict.get(descr.getHandle());
					if (keys == null) {
						keys = new HashSet();
						dict.put(descr.getHandle(), keys);
					}

					for (String ppk : ppks) {
						if (!ppk.isEmpty()) {
							keys.add(ppk);
						}
					}
				}
			}
		}
	}

	/**
	 * Calculate the match probabilities for all words in the given node
	 * descriptor and the word given in s.
	 *
	 * @param matches the list to which the resulting DictionaryMatch instances
	 * will be added.
	 * @param descriptor is the node descriptor whose node names should be
	 * examined.
	 * @param s is the word the words in this list should be compared to.
	 * @param pThresh is the minimum probability for which a match is emitted.
	 */
	static protected void fillMatchesForNode(List<DictionaryMatch> matches,
			NodeDescriptor descriptor, String s, double pThresh) {
		// Fill the dictionary with the names present in this node
		final NodeHandle handle = descriptor.getHandle();
		for (NodeName nameEntry : descriptor.getNames()) {
			// Create a list that contains the temporary matches for a single
			// name
			List<DictionaryMatch> nameMatches = new ArrayList();

			// Normalize the given word according to the comparison mode of the
			// name entry, check whether the result is valid
			final ComparisonMode mode = nameEntry.getComparisonMode();
			final String wComp = DictionaryAlgorithms.normalizeWord(s, mode);
			if (wComp == null || wComp.isEmpty()) {
				continue;
			}

			// Fetch all words for the given name entry
			final String name = nameEntry.getName();
			final Set<WordPosition> wordPositions = DictionaryAlgorithms.getWords(name);
			for (WordPosition wordPosition : wordPositions) {
				// Extract the current word
				final int begin = wordPosition.getBegin();
				final int end = wordPosition.getEnd();
				final String w = wordPosition.getWord();

				// Normalize both words and make sure the normalization result
				// is still a valid word
				final String wName = DictionaryAlgorithms.normalizeWord(w, mode);
				if (wName == null || wName.isEmpty()) {
					continue;
				}

				// Depending on the comparison mode add a match
				switch (mode) {
					case FUZZY:
						final double p = DictionaryAlgorithms.matchProbability(wName, wComp);
						if (p > MIN_MATCH_PROBABILITY) {
							nameMatches.add(new DictionaryMatch(handle, nameEntry,
									begin, end, p));
						}
						break;
					case CLEAN:
					case SOFT:
						if (wComp.equals(wName)) {
							nameMatches.add(new DictionaryMatch(handle, nameEntry,
									begin, end, 1.0));
						}
						break;
					case LOWERCASE:
					case STRICT:
						// Generate a regular expression for matching
						final String patternStr = "^[^\\w]*(" + Pattern.quote(wName) + ")[^\\w]*$";
						final Pattern pattern = Pattern.compile(patternStr);
						final Matcher matcher = pattern.matcher(wComp);
						if (matcher.matches()) {
							nameMatches.add(new DictionaryMatch(handle, nameEntry,
									begin, end, 1.0,
									new LineSegment(matcher.start(1), matcher.end(1))));
						}
						break;
				}
			}

			// Add all name matches to the output list, remove overlapping
			// matches that refer to the same word but differ in non-alphabetic
			// characters
			matches.addAll(DictionaryMatch.reduceMatches(name, nameMatches));
		}
	}

	/**
	 * Queries the database for a collection of words. If the word (or any
	 * related word) is contained within the ontology database, the result will
	 * contain a mapping between the queried word and the corresponding ontology
	 * dictionary entries.
	 *
	 * @param words is a list of AssociativeWord instances. A AssociativeWord
	 * consists of an arbitrary string and a user-defined corresponding object.
	 * This way the caller of this function can attach arbitrary information to
	 * the string to be sought for.
	 * @return a mapping between AssociativeWord instances and the found
	 * DictionaryLink instances. Note that words that were not found or that
	 * turned out to be "stop words" are not returned.
	 */
	public Map<String, Collection<DictionaryMatch>> query(Collection<String> words) {
		// Calculate the reduced variants of each given word and remember the
		// relationship in a hash map
		Map<String, Collection<String>> keyWordMap = buildKeyWordMap(words);

		// Perform the actual query, obtain a mapping between the queried keys
		// and linked nodes.
		Map<String, Collection<NodeDescriptor>> queryRes
				= driver.queryDictionary(keyWordMap.keySet());

		// Iterate over all found keys and build the result
		Map<String, Collection<DictionaryMatch>> res = new HashMap();
		for (String key : queryRes.keySet()) {
			// Fetch the links for the word
			Collection<NodeDescriptor> descriptors = queryRes.get(key);

			// Build the matches for each word, sort by match probability
			for (String word : keyWordMap.get(key)) {
				Collection<DictionaryMatch> matches = res.get(word);
				if (matches == null) {
					matches = new LinkedList();
					res.put(word, matches);
				}
				for (NodeDescriptor descriptor : descriptors) {
					fillMatchesForNode((List) matches, descriptor, word, MIN_MATCH_PROBABILITY);
				}
			}
		}
		return res;
	}

	/**
	 * Updates the given node in the dictionary -- deletes all old dictionary
	 * entries for that node and writes new dictionary entries.
	 *
	 * @param node is the node for which the dictionary entries will be updated.
	 * If the node does not exist in the database, nothing will happen.
	 */
	public void updateNode(NodeDescriptor node) {
		// Delete any old remains of the given node
		driver.deleteDictionaryEntry(node.getHandle());

		// Make sure the node exists
		if (driver.queryNodeExists(node.getHandle())) {
			// Fetch all keys belonging to the node and store them -- do not call
			// the store function if the key set is empty (as we have nothing to
			// store)
			Map<NodeHandle, Set<String>> keys = new HashMap();
			fillDictionaryEntriesForNode(node, keys);
			if (!keys.isEmpty()) {
				driver.storeDictionaryEntries(keys);
			}
		} else {
			throw new InvalidNodeException("The given node does not exist in the"
					+ "database and cannot be updated!");
		}
	}

}
