/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.util.LineSegment;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import java.util.HashMap;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

/**
 * Collection of static algorithms used throughout the ontology dictionary. This
 * class provides a central place in which all external dependencies and
 * advanced algorithms used in the abstract ontology database are bundled.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public final class DictionaryAlgorithms {

	/**
	 * Minimim word length.
	 */
	protected static final int MIN_WORD_LENGTH = 3;

	/**
	 * Pattern containing all stop-words. A regular expression was used as these
	 * are compiled to an efficient state machine.
	 */
	protected static final Pattern STOPWORD = Pattern.compile(
		"(in|if|is|the|then|this|thus|a|an|and|but|when|where|whom|"
		+ "what|why|we|so|of|on|et|al)");

	/**
	 * Pattern containing all suffixes which should be removed. A regular
	 * expression was used as these are compiled to an efficient state machine.
	 */
	protected static final Pattern SUFFIXES = Pattern.compile(
		"(illa|ella|" // Four characters
		+ "ous|ing|ish|" // Three characters
		+ "on|os|us|um|ii|is|ae|as|ei|ea|es|la|si|" // Latin two characters
		+ "ax|my|" // Greek two characters
		+ "th|ny|ed|ly|er|al|" // English
		+ "y|a|e|s|u" // One character
		+ ")$");

	/**
	 * Pattern used to extract words from a string. Words may only consist of
	 * alpha-numerical characters.
	 */
	protected static final Pattern WORD = Pattern.compile("[\\w0-9]+");

	/**
	 * Pattern used to extract words from a string. This strict pattern only
	 * splits words at whitespace.
	 */
	protected static final Pattern WORD_STRICT = Pattern.compile("[^\\s]+");

	/**
	 * Alphanumeric character.
	 */
	protected static final Pattern ALPHANUMERIC_CHARACTER = Pattern.compile("[\\w0-9]");

	/**
	 * Returns true if the given string contains at least one alphanumeric
	 * character.
	 *
	 * @param s is the string which should be checked.
	 *
	 * @return true if the given string contains an alphanumeric character.
	 */
	public static boolean containsAlphanumericChar(String s) {
		return ALPHANUMERIC_CHARACTER.matcher(s).find();
	}

	public static boolean isAlphabetic(char c) {
		return ((c >= 'A') && (c <= 'Z')) || ((c >= 'a') && (c <= 'z'));
	}

	public static boolean isNumeric(char c) {
		return (c >= '0') && (c <= '9');
	}

	public static boolean isAlphanumeric(char c) {
		return isAlphabetic(c) || isNumeric(c);
	}

	/**
	 * Convert the incomming string to lowercase, remove any special character.
	 *
	 * @param s is the string from which all non-alphabetic characters should be
	 * stripped.
	 *
	 * @return the prepared version.
	 */
	public static String stripNonAlphabetic(String s) {
		// Strip all non-letter characters
		final char[] chars = s.toCharArray();
		final StringBuilder builder = new StringBuilder(chars.length);
		for (int i = 0; i < chars.length; i++) {
			if (isAlphabetic(chars[i])) {
				// "| 0x20" is guaranteed to transform the letters in the range
				// above to lower case.
				builder.append((char) (chars[i] | 0x20));
			}
		}

		// Fetch the resulting string, check whether the word is a stop word.
		return builder.toString();
	}

	/**
	 * Convert the incomming string to lowercase, remove any non alphanumeric
	 * character.
	 *
	 * @param s is the string from which all non-alphanumeric characters should
	 * be stripped.
	 *
	 * @return the prepared version.
	 */
	public static String stripNonAlphanumeric(String s) {
		// Strip all non-letter characters
		final char[] chars = s.toCharArray();
		final StringBuilder builder = new StringBuilder(chars.length);
		for (int i = 0; i < chars.length; i++) {
			if (isAlphabetic(chars[i])) {
				// "| 0x20" is guaranteed to transform the letters in the range
				// above to lower case.
				builder.append((char) (chars[i] | 0x20));
			} else if (isNumeric(chars[i])) {
				builder.append(chars[i]);
			}
		}

		// Fetch the resulting string, check whether the word is a stop word.
		return builder.toString();
	}

	/**
	 * Strip all whitespace from the given string.
	 *
	 * @param s is the string from which all whitespace should be stripped.
	 *
	 * @return the prepared version.
	 */
	public static String stripWhitespace(String s) {
		// Strip all non-letter characters
		final char[] chars = s.toCharArray();
		final StringBuilder builder = new StringBuilder(chars.length);
		for (int i = 0; i < chars.length; i++) {
			if (!Character.isWhitespace(chars[i])) {
				builder.append(chars[i]);
			}
		}

		// Fetch the resulting string, check whether the word is a stop word.
		return builder.toString();
	}

	/**
	 * Returns the position of the first whitespace in the given string or
	 * Integer.MAX_VALUE if there is no whitespace in the string.
	 *
	 * @param s is the string of which the first whitespace position should be
	 * returned.
	 *
	 * @return the position of the first whitepsace character or
	 * Integer.MAX_VALUE if the string does not contain whitespace.
	 */
	public static int firstWhitespaceIndex(String s) {
		final char[] chars = s.toCharArray();
		for (int i = 0; i < chars.length; i++) {
			if (Character.isWhitespace(chars[i])) {
				return i;
			}
		}
		return Integer.MAX_VALUE;
	}

	/**
	 * The WordPosition class is used to classify a word position within a
	 * NodeName instance.
	 */
	public static final class WordPosition extends LineSegment
		implements Comparable<WordPosition> {

		/**
		 * The string this word position refers to.
		 */
		public final String s;

		/**
		 * Constructor of the WordPosition class.
		 *
		 * @param s is string this word position refers to.
		 * @param begin is the start character of the word.
		 * @param end is the end character of the word.
		 */
		public WordPosition(String s, int begin, int end) {
			super(begin, end);
			this.s = s;
		}

		/**
		 * Constructor of the WordPosition class.
		 *
		 * @param begin is the start character of the word.
		 * @param end is the end character of the word.
		 */
		public WordPosition(int begin, int end) {
			super(begin, end);
			this.s = null;
		}

		/**
		 * Returns the word this WordPosition refers to.
		 *
		 * @return the word this WordPosition refers to.
		 */
		public String getWord() {
			if (s != null) {
				return s.substring(begin, end);
			} else {
				return "";
			}
		}

		/**
		 * Two WordPosition instances equal if they point to the same string and
		 * have the same begin and end.
		 *
		 * @param obj is the object this WordPosition should be compared to.
		 *
		 * @return true if the two WordPosition instances equal, false
		 * otherwise.
		 */
		@Override
		public boolean equals(Object obj) {
			if (obj == null || obj.getClass() != getClass()) {
				return false;
			}

			WordPosition pos = (WordPosition) obj;
			return pos.begin == begin && pos.end == end;
		}

		/**
		 * Returns the hash code which is constructed from the begin, the end
		 * position and the hash code of the string this name refers to.
		 *
		 * @return the hash code.
		 */
		@Override
		public int hashCode() {
			return 117 * begin + 53 * end;
		}
		
		

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int compareTo(WordPosition o) {
		if (o == null) {
			return 1;
		}
		final int beginCompare = Integer.compare(this.begin, o.getBegin());
		if (beginCompare == 0) {
			return Integer.compare(this.end, o.getEnd());
		}
		return beginCompare;
	}

	}

	/**
	 * Adds word positions for all matches of the given regular expression to
	 * the given set.
	 *
	 * @param words is a set of word positions to which the matches for the
	 * given pattern in the given string should be added.
	 * @param s is the string in which the matches should be searched.
	 * @param p is the pattern for which the match positions should be searched.
	 */
	private static void addMatchPositions(Set<WordPosition> words, String s,
		Pattern p) {
		Matcher m = p.matcher(s);
		while (m.find()) {
			if (containsAlphanumericChar(m.group(0))) {
				words.add(new WordPosition(s, m.start(), m.end()));
			}
		}
	}

	/**
	 * As the getWords function is a linear time algorithm in the text length we
	 * do not want to recalculate that all the time. Thus we implement a cache
	 * for that.
	 *
	 * We use Object as key here to enforce the usage of objects hashCode
	 * method, which should be faster than a string hashCode and enforces the
	 * actual same object.
	 */
	private static HashMap<Object, Set<WordPosition>> wordsCache = new HashMap<>();
	private static HashMap<Object, Set<WordPosition>> strictWordsCache = new HashMap<>();

	/**
	 * Returns a list of words from the given string.
	 *
	 * @param s is the words for which the words should be calculated.
	 * @param onlyStrict if true, only words seperated by whitespace are
	 * returned, otherwise both words separated by whitespace and words
	 * separated by special characters are returned.
	 *
	 * @return a list of words within the name
	 */
	public static Set<WordPosition> getWords(String s, boolean onlyStrict) {
		//if possible, retrieve the result of this algorithm from our caches,
		//either for the strict or the regular case.
		if (onlyStrict) {
			if (strictWordsCache.containsKey(s)) {
				return strictWordsCache.get(s);
			}
		} else {
			if (wordsCache.containsKey(s)) {
				return wordsCache.get(s);
			}
		}
		//otherwise retrieve the words.
		final Set<WordPosition> words = new TreeSet();
		addMatchPositions(words, s, WORD_STRICT);
		if (!onlyStrict) {
			addMatchPositions(words, s, WORD);
		}
		//and store the result in caches.
		if (onlyStrict) {
			strictWordsCache.put(s, words);
		} else {
			wordsCache.put(s, words);
		}
		return words;
	}

	/**
	 * This clears the caches for the getWords method.
	 */
	public static void purgeWordsCache() {
		wordsCache.clear();
		strictWordsCache.clear();
	}

	/**
	 * Returns a list of words from the given string (both strict and non-strict
	 * matches).
	 *
	 * @param s is the words for which the words should be calculated.
	 *
	 * @return a list of words within the name
	 */
	public static Set<WordPosition> getWords(String s) {
		return getWords(s, false);
	}

	/**
	 * Normalizes the node name according to the given comparison mode.
	 *
	 * @param s is the input string (should be a single word).
	 * @param mode is the comparison mode for which the node name should be
	 * normalized.
	 *
	 * @return the normalized word
	 */
	public static String normalizeWord(String s, ComparisonMode mode) {
		switch (mode) {
			case FUZZY:
				s = stripNonAlphabetic(s);
				return isStopWord(s) ? null : s;
			case SOFT:
				return stripNonAlphabetic(s);
			case CLEAN:
				return stripNonAlphanumeric(s);
			case LOWERCASE:
				return s.toLowerCase();
			case STRICT:
				return s;
			default:
				return null;
		}
	}

	/**
	 * Returns true if the given word is a stop word. Note that the word that
	 * should be checked already has to be converted to lowercase letters.
	 *
	 * @param word to be checked, must be lowercase.
	 *
	 * @return true if the word is a stop word.
	 */
	public static boolean isStopWord(String word) {
		return word.length() < MIN_WORD_LENGTH || STOPWORD.matcher(word).matches();
	}

	/**
	 * Extracts the word stem of a certain word. The word has to be in lowercase
	 * and free of punctuation symbols.
	 *
	 * @param word is the word of which the word stem should be calculated.
	 *
	 * @return the actual word stem, or null if the given word was empty
	 */
	public static String wordStem(String word) {
		if (word == null || word.isEmpty()) {
			return null;
		}

		while (true) {
			final Matcher matcher = SUFFIXES.matcher(word);
			final String stem = matcher.replaceAll("");
			if (word.equals(stem) || stem.length() < MIN_WORD_LENGTH) {
				break;
			}
			word = stem;
		}
		return word;
	}

	/**
	 * Calculates the edit distance between word w1 and word w2. Currently the
	 * Levensthein distance is used.
	 *
	 * @param w1 is the first word.
	 * @param w2 is the second word.
	 *
	 * @return the count of steps needed to transform w1 to w2.
	 */
	public static int editDistance(String w1, String w2) {
		return StringUtils.getLevenshteinDistance(w1, w2);
	}

	/**
	 * Calculates the "match probability" between w1 and w2 as a value between
	 * 0.0f and 1.0f. Basically this is one minus the edit distance divided by
	 * the length of the longest of the two words.
	 *
	 * @param w1 is the first word.
	 * @param w2 is the second word.
	 *
	 * @return a probability value describing how probable it is for w1 to match
	 * w2.
	 */
	public static float matchProbability(String w1, String w2) {
		return Math.max(0.0f, 1.0f - (float) editDistance(w1, w2)
			/ (float) Math.max(w1.length(), w2.length()));
	}

	/**
	 * Function that can be used to calculated dimensionalty reduced keys for
	 * the given word.
	 *
	 * @param word is word for which the keys shall be calculated.
	 * @param targetMode specifies the target mode for which the keys are
	 * generated. Only keys for modes more general or equally general as the
	 * given mode are generated.
	 *
	 * @return an array containing the resulting keys.
	 */
	public static String[] calculateWordKeys(String word, ComparisonMode targetMode) {
		// If the given string is empty, return an empty array.
		if ((word == null) || word.isEmpty()) {
			return new String[]{};
		}

		// Calculate the most general version of the word
		final ComparisonMode modes[] = new ComparisonMode[]{
			ComparisonMode.FUZZY, ComparisonMode.SOFT, ComparisonMode.LOWERCASE
		};
		for (ComparisonMode mode : modes) {
			// Only generate the keys for more ore equally general (expressed by
			// the value) comparison modes
			if (mode.getValue() > targetMode.getValue()) {
				break;
			}

			String generalWord = normalizeWord(word, mode);

			// Special treatment for the FUZZY mode -- calculate the word stem
			// here
			if (mode == ComparisonMode.FUZZY) {
				generalWord = wordStem(generalWord);
			}

			if (generalWord != null && !generalWord.isEmpty()) {
				return new String[]{generalWord};
			}
		}

		// If the given word could not be generalized, return an empty array
		return new String[]{};
	}

	/**
	 * Function that can be used to calculated dimensionalty reduced keys for
	 * the given word with "strict" as target comparison mode.
	 *
	 * @param word is word for which the keys shall be calculated.
	 *
	 * @return an array containing the resulting keys.
	 */
	public static String[] calculateWordKeys(String word) {
		return calculateWordKeys(word, ComparisonMode.STRICT);
	}

}
