/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.util.LineSegment;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;

/**
 * The DictionaryMatch represents a match between a single word in a NodeName
 * instance together with its match probability.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public class DictionaryMatch extends LineSegment implements Comparable<DictionaryMatch> {

	/**
	 * Reference to the ontology node.
	 */
	private final NodeHandle node;

	/**
	 * Reference to the node name instance of the given node the match refers
	 * to.
	 */
	private final NodeName name;

	/**
	 * The match probability between the two words.
	 */
	private double probability;

	/**
	 * If set, this DictionaryMatch refers to a partial match in the word it
	 * refers to.
	 */
	private LineSegment partialMatch;

	/**
	 * Removes all overlapping matches to the same node name that only differ in
	 * non-alphanumeric characters.
	 *
	 * @param underlyingText specifies the text the positions in the dicrionary
	 * match refer to.
	 * @param matches is the list of matches that will be reduced.
	 * @return the resulting reduced match list.
	 */
	public static Collection<DictionaryMatch> reduceMatches(String underlyingText,
			List<DictionaryMatch> matches) {
		// Count the matches and create an array that indicates whether a
		// certain match is invalid, because a better match exists at the same
		// position
		final int matchCount = matches.size();
		final boolean invalidMatch[] = new boolean[matches.size()];
		for (int i = 0; i < matchCount; i++) {
			invalidMatch[i] = false;
		}

		// Iterate over all pairs {i, j} and compare them. Mark overlapping
		// matches which only differ in whitespace/non-alphabetic characters for
		// which a better match exists as "invalid".
		for (int i = 0; i < matchCount; i++) {
			// If this match was marked as "invalid" continue
			if (invalidMatch[i]) {
				continue;
			}

			DictionaryMatch mI = matches.get(i);
			int curIdx = i;
			for (int j = i + 1; j < matchCount; j++) {
				DictionaryMatch mJ = matches.get(j);

				// Check whether the two matches overlap and only differ in
				// whitespace
				LineSegment intersection = LineSegment.intersection(mI, mJ);
				LineSegment union = LineSegment.union(mI, mJ);
				if (intersection == null) {
					continue;
				}

				// Get the substring at the begin and end of the word
				final String s1 = underlyingText.substring(union.getBegin(),
						intersection.getBegin()); // prefix
				final String s2 = underlyingText.substring(intersection.getEnd(),
						union.getEnd()); // suffix
				final boolean s1f = DictionaryAlgorithms.containsAlphanumericChar(s1);
				final boolean s2f = DictionaryAlgorithms.containsAlphanumericChar(s2);
				if (s1f || s2f) {
					// The two matches differ in the word they cover and not only
					// in whitespace/non-alphanumeric characters
					continue;
				}

				// Check which of the two matches is more probable -- for equal
				// probability select the shorter match
				final double EPSILON = 0.000001;
				final double pI = mI.getProbability();
				final double pJ = mJ.getProbability();
				int winner = -1; // -1 refers to mI
				if (Math.abs(pI - pJ) < EPSILON) {
					// The two probabilities are equal select the shorter one or
					// the first one
					if (mI.length() > mJ.length()) {
						winner = 1; // mJ wins
					}
				} else if (pJ > pI) {
					winner = 1; // mJ wins
				}
				if (winner == -1) {
					invalidMatch[j] = true;
				} else {
					invalidMatch[curIdx] = true;
					curIdx = j;
					mI = mJ;
				}
			}
		}

		// Add all valid matches to the result
		Collection<DictionaryMatch> res = new ArrayList();
		for (int i = 0; i < matchCount; i++) {
			if (!invalidMatch[i]) {
				res.add(matches.get(i));
			}
		}
		return res;
	}

	/**
	 * Creates a new instance of DictionaryMatch.
	 *
	 * @param node is a reference to the node the match belongs to.
	 * @param name is a reference to the NodeName instance this match refers to.
	 * @param begin marks the beginning of the substring in the name this match
	 * refers to.
	 * @param end marks the end of the substring in the name this match refers
	 * to.
	 * @param probability is the match probability.
	 * @param partialMatch if set, this indicates that the dictionary match
	 * refers to a partial word in the reference text. May also be NULL.
	 */
	public DictionaryMatch(NodeHandle node, NodeName name, int begin, int end,
			double probability, LineSegment partialMatch) {
		this.node = node;
		this.name = name;
		this.begin = begin;
		this.end = end;
		this.probability = probability;
		this.partialMatch = partialMatch;
	}

	/**
	 * Creates a new instance of DictionaryMatch.
	 *
	 * @param node is a reference to the node the match belongs to.
	 * @param name is a reference to the NodeName instance this match refers to.
	 * @param begin marks the beginning of the substring in the name this match
	 * refers to.
	 * @param end marks the end of the substring in the name this match refers
	 * to.
	 * @param probability is the match probability.
	 */
	public DictionaryMatch(NodeHandle node, NodeName name, int begin, int end,
			double probability) {
		this(node, name, begin, end, probability, null);
	}

	public DictionaryMatch(NodeHandle node, NodeName name) {
		this(node, name, Integer.MAX_VALUE, Integer.MIN_VALUE, 0.0);
	}

	public DictionaryMatch(int begin, int end) {
		this(null, null, begin, end, 0.0);
	}

	/**
	 * Returns the reference to the related node.
	 *
	 * @return the node that is associated to this entry.
	 */
	public NodeHandle getNode() {
		return node;
	}

	/**
	 * Returns the reference to the NodeName this match refers to.
	 *
	 * @return the reference to the NodeName this match refers to.
	 */
	public NodeName getNameDescr() {
		return name;
	}

	/**
	 * Returns the name of the NodeName this match refers to.
	 *
	 * @return the name of the NodeName this match refers to or an empty string
	 * if this match is associated to the name "null".
	 */
	public String getName() {
		return (name == null) ? "" : name.getName();
	}

	/**
	 * Returns the match probability as a value between 0.0f and 1.0f
	 *
	 * @return the match probability.
	 */
	public double getProbability() {
		return probability;
	}

	/**
	 * Sets the probability to the given value.
	 *
	 * @param probability
	 */
	public void setProbability(double probability) {
		this.probability = probability;
	}

	/**
	 * Returns the relative match probability (relative to the length of the
	 * complete word).
	 *
	 * @param maxCharCount is the maximum character count that should be used
	 * for the calculation.
	 * @return the relative match probability.
	 */
	public double getRelativeMatchProbability(int maxCharCount) {
		return probability * (double) (end - begin) / (double) maxCharCount;
	}

	/**
	 * Returns the partial match position (if any).
	 *
	 * @return the partial match this match refers to or null if none.
	 */
	public LineSegment getPartialMatch() {
		return partialMatch;
	}

	/**
	 * Allows to sort the matches by their weight.
	 *
	 * @param t the match this match should be compared to.
	 * @return the comparison result.
	 */
	@Override
	public int compareTo(DictionaryMatch t) {
		return (probability > t.probability) ? 1
				: ((probability < t.probability) ? -1 : 0);
	}

	/**
	 * Checks whether the two DictionaryMatch instances equal. This is the case
	 * if they refer to the same node, name and the begin and end are equal.
	 *
	 * @param obj is the object this match should be compared to.
	 * @return true if the two objects are the same, false otherwise.
	 */
	@Override
	public boolean equals(Object obj) {
		if ((obj == null) || (obj.getClass() != getClass())) {
			return false;
		}

		DictionaryMatch m = (DictionaryMatch) obj;
		return (begin == m.begin) && (end == m.end)
				&& (Objects.equals(node, m.node))
				&& (Objects.equals(name, m.name));
	}

	/**
	 * Returns a hash code of the object that depends on the node handle, the
	 * node name and the begin and end.
	 *
	 * @return the hash code.
	 */
	@Override
	public int hashCode() {
		int hash = 5;
		hash = 53 * hash + Objects.hashCode(node);
		hash = 53 * hash + Objects.hashCode(name);
		hash = 53 * hash + begin;
		hash = 53 * hash + end;
		return hash;
	}
}
