/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.driver.Association;
import de.unibi.techfak.scie.ontology.driver.AssociationDirection;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Stack;

/**
 * The Node class extends NodeDescriptor and adds convenience functions that
 * allow to list all associated nodes or to simply add new associated nodes.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class Node extends NodeDescriptor {

	/**
	 * Reference to an NodeFactory -- the node factory is used to convert (lists
	 * of) NodeHandles and NodeDescriptors to Node instances.
	 */
	private final NodeFactory factory;
	/**
	 * Reference to the database driver (as obtained from the factory instance).
	 */
	private final OntologyDatabaseDriver driver;

	/**
	 * Constructor of the Node class.
	 *
	 * @param factory is an instance of the NodeFactory class that should be
	 * used for lookups.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 * @param root specifies whether the node is a root node.
	 */
	Node(NodeFactory factory, NodeHandle handle, Collection<NodeName> names,
			boolean root) {
		super(handle, names, root);
		this.factory = factory;
		this.driver = factory.getDriver();
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param factory is an instance of the NodeFactory class that should be
	 * used for lookups.
	 * @param node is a reference to the node descriptor from which this
	 * instance should be derived.
	 */
	Node(NodeFactory factory, NodeDescriptor node) {
		this(factory, node.getHandle(), node.getNames(), node.isRoot());
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle,
			Collection<NodeName> names, boolean root) {
		this(db.getNodeFactory(), handle, names, root);
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle, NodeName name,
			boolean root) {
		this(db, handle, Arrays.asList(new NodeName[]{name}), root);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle,
			Collection<NodeName> names) {
		this(db, handle, names, false);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle,
			NodeName name) {
		this(db, handle, name, false);
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single name that should be attached to the node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle, String name,
			boolean root) {
		this(db, handle, new NodeName(name), root);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single name that should be attached to the node.
	 */
	public Node(OntologyDatabase db, NodeHandle handle, String name) {
		this(db, handle, new NodeName(name), false);
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			Collection<NodeName> names, boolean root) {
		this(db, new NodeHandle(handle, id), names, root);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			Collection<NodeName> names) {
		this(db, handle, id, names, false);
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			NodeName name, boolean root) {
		this(db, new NodeHandle(handle, id), name, root);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			NodeName name) {
		this(db, handle, id, name, false);
	}

	/**
	 * Constructor of the Node class.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single name that should be attached to the node.
	 * @param root specifies whether the node is a root node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			String name, boolean root) {
		this(db, handle, id, new NodeName(name), root);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single name that should be attached to the node.
	 */
	public Node(OntologyDatabase db, OntologyHandle handle, int id,
			String name) {
		this(db, handle, id, new NodeName(name), false);
	}

	/**
	 * Constructor of the Node class. The node is not a root node.
	 *
	 * @param db is a reference to the ontology database.
	 * @param node is a reference to the node descriptor from which this
	 * instance should be derived.
	 */
	public Node(OntologyDatabase db, NodeDescriptor node) {
		this(db, node.getHandle(), node.getNames(), node.isRoot());
	}

	/**
	 * Return a collection of nodes that are associated to this node.
	 *
	 * @param associationTypes is a bit mask (values of the AssociationType enum
	 * combined with a bitwise or) specifying the association types that should
	 * be returned.
	 * @param direction is the direction for which associations should be
	 * returned.
	 * @return a list of associated nodes or an empty list if no associations
	 * with the given constraints were found.
	 */
	public Collection<Node> getAssociatedNodes(int associationTypes,
			AssociationDirection direction) {
		Collection<Association> associations = driver.getAssociations(
				handle, associationTypes, direction);
		Collection<NodeHandle> res = new ArrayList(associations.size());
		for (Association association : associations) {
			if (association.getSource().equals(handle)) {
				res.add(association.getTarget());
			} else {
				res.add(association.getSource());
			}
		}
		return factory.handlesToNodeList(res);
	}

	/**
	 * Returns a list of associated nodes with all types and regardless of the
	 * association direction.
	 *
	 * @return a list of associated nodes or an empty list if no associations
	 * with the given constraints were found.
	 */
	public Collection<Node> getAssociatedNodes() {
		return getAssociatedNodes(AssociationType.ALL, AssociationDirection.BOTH);
	}

	/**
	 * Return a collection of nodes that are associated to this node, for both
	 * association directions.
	 *
	 * @param associationTypes is a bit mask (values of the AssociationType enum
	 * combined with a bitwise or) specifying the association types that should
	 * be returned.
	 * @return a list of associated nodes or an empty list if no associations
	 * with the given constraints were found.
	 */
	public Collection<Node> getAssociatedNodes(int associationTypes) {
		return getAssociatedNodes(associationTypes, AssociationDirection.BOTH);
	}

	/**
	 * Return a collection of nodes that are associated to this node, regardless
	 * of the association type.
	 *
	 * @param direction is the direction for which associations should be
	 * returned.
	 * @return a list of associated nodes or an empty list if no associations
	 * with the given constraints were found.
	 */
	public Collection<Node> getAssociatedNodes(AssociationDirection direction) {
		return getAssociatedNodes(AssociationType.ALL, direction);
	}

	/**
	 * This checks whether there exists a list of nodes this, n_1, n_2, ...,
	 * n_k, p such that there exists an ISA-Association from each node in the
	 * list to the next one. In other words: This method returns true if this
	 * node is an (in)direct child of p.
	 *
	 * @param supposedParent The supposed parent p.
	 * @return true if this node is an (in)direct child of p.
	 */
	public boolean isA(NodeHandle supposedParent) {
		//do a depth-first-search to find parents.
		final Stack<Node> stack = new Stack<>();
		stack.push(this);

		final int associationType = AssociationType.IS_A.getValue();

		while (!stack.empty()) {
			final Node currentNode = stack.pop();
			final Collection<Node> parents = currentNode.getAssociatedNodes(
					associationType, AssociationDirection.OUTBOUND);
			for (final Node parent : parents) {
				if (parent.getId() == supposedParent.getId()) {
					//if we found our supposed parent, we are finished.
					return true;
				} else {
					//otherwise we push that parent on the stack.
					stack.push(parent);
				}
			}
		}
		//If we did not find the supposed parent, return false.
		return false;
	}

	/**
	 * Adds a new association to the given node. Note that the node has to be in
	 * the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (target node,
	 * outbound association).
	 * @param type is the type of the new association.
	 */
	public void addAssociationTo(NodeHandle node, AssociationType type) {
		driver.storeAssociation(new Association(handle, node, type));
	}

	/**
	 * Adds a new association to the given node. Note that the node has to be in
	 * the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (target node,
	 * outbound association).
	 */
	public void addAssociationTo(NodeHandle node) {
		addAssociationTo(node, AssociationType.DEFAULT);
	}

	/**
	 * Adds a new association to the given node. Note that the node has to be in
	 * the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (target node,
	 * outbound association).
	 * @param type is the type of the new association.
	 */
	public void addAssociationTo(NodeDescriptor node, AssociationType type) {
		addAssociationTo(node.getHandle(), type);
	}

	/**
	 * Adds a new association to the given node. Note that the node has to be in
	 * the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (target node,
	 * outbound association).
	 */
	public void addAssociationTo(NodeDescriptor node) {
		addAssociationTo(node.getHandle());
	}

	/**
	 * Adds a new association from the given node. Note that the node has to be
	 * in the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (source node,
	 * inbound association).
	 * @param type is the type of the new association.
	 */
	public void addAssociationFrom(NodeHandle node, AssociationType type) {
		driver.storeAssociation(new Association(node, handle, type));
	}

	/**
	 * Adds a new association from the given node. Note that the node has to be
	 * in the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (source node,
	 * inbound association).
	 */
	public void addAssociationFrom(NodeHandle node) {
		addAssociationFrom(node, AssociationType.DEFAULT);
	}

	/**
	 * Adds a new association from the given node. Note that the node has to be
	 * in the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (source node,
	 * inbound association).
	 * @param type is the type of the new association.
	 */
	public void addAssociationFrom(NodeDescriptor node, AssociationType type) {
		addAssociationFrom(node.getHandle(), type);
	}

	/**
	 * Adds a new association from the given node. Note that the node has to be
	 * in the same ontology as this node.
	 *
	 * @param node is the node the association should be added to (source node,
	 * inbound association).
	 */
	public void addAssociationFrom(NodeDescriptor node) {
		addAssociationFrom(node.getHandle());
	}

	/**
	 * Deletes any present association to the given node (outbound, this node is
	 * the source node). Nothing happens if no such association exists.
	 *
	 * @param node is the node to which all associations should be removed.
	 */
	public void deleteAssociationTo(NodeHandle node) {
		driver.deleteAssociation(new Association(this.handle, node));
	}

	/**
	 * Deletes any present association to the given node (outbound, this node is
	 * the source node). Nothing happens if no such association exists.
	 *
	 * @param node is the node to which all associations should be removed.
	 */
	public void deleteAssociationTo(NodeDescriptor node) {
		deleteAssociationTo(node.getHandle());
	}

	/**
	 * Deletes any present association frmo the given node (inbound, this node
	 * is the target node). Nothing happens if no such association exists.
	 *
	 * @param node is the node to which all associations should be removed.
	 */
	public void deleteAssociationFrom(NodeHandle node) {
		driver.deleteAssociation(new Association(node, this.handle));
	}

	/**
	 * Deletes any present association frmo the given node (inbound, this node
	 * is the target node). Nothing happens if no such association exists.
	 *
	 * @param node is the node to which all associations should be removed.
	 */
	public void deleteAssociationFrom(NodeDescriptor node) {
		deleteAssociationFrom(node.getHandle());
	}
}
