/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.dictionary_name_finder;

import de.unibi.techfak.scie.util.LineSegment;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms.WordPosition;
import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

/**
 * DictionaryNameFinder is used to finde complete NodeName instances in a text.
 * Interesting words are added to the finder using the "process
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class DictionaryNameFinder {

	/**
	 * Specifies the count of words that are processed in a single batch.
	 */
	private static final int MIN_WORD_COUNT = 100;

	/**
	 * The wordGroupMap is used to sort the dictionary matches into the so
	 * called word group maps.
	 */
	private final WordGroupMap wordGroupMap = new WordGroupMap();

	/**
	 * Reference to the ontology database.
	 */
	private final OntologyDatabase db;

	/**
	 * Reference to the document text.
	 */
	private final String documentText;

	/**
	 * Probability threshold.
	 */
	private final double pThresh;

	/**
	 * The syntactic model that should be used for deriving the word references.
	 */
	private final SyntacticProbabilityModel model;

	/**
	 * A map between the words that should be queried and all positions in the
	 * text where the word occured.
	 */
	private final Map<String, Collection<WordPosition>> queryBuffer;

	/**
	 * Queries the given words from the ontology database and creates
	 * OntologyWordAnnotation instances for each match that is found.
	 */
	private void flushQueryBuffer() {
		// Query the database for the given words
		Map<String, Collection<DictionaryMatch>> res = db.query(queryBuffer.keySet());
		for (String key : res.keySet()) {
			// Fetch all the positions that correspond to the given key
			Collection<WordPosition> positions = queryBuffer.get(key);
			if (positions == null) {
				continue;
			}

			// Insert a word reference for each position and each match
			for (WordPosition position : positions) {
				Collection<DictionaryMatch> matches = res.get(key);
				for (DictionaryMatch match : matches) {
					LineSegment partialMatch = match.getPartialMatch();
					int offsBegin = (partialMatch == null) ? 0 : partialMatch.getBegin();
					int offsEnd = (partialMatch == null) ? 0 : position.length() - partialMatch.getEnd();
					final WordReference wordRef = new WordReference(
							position.getBegin() + offsBegin, position.getEnd() - offsEnd, match);
					wordGroupMap.insertWord(wordRef);
				}
			}
		}

		// Clear the query buffer
		queryBuffer.clear();
	}

	/**
	 * Creates a new instance of the DictionaryNameFinder class and connects it
	 * to the given database and the given probability model.
	 *
	 * @param db is a reference to the ontology database that should be used for
	 * querying the ontology names.
	 * @param documentText is a reference to the underlying document text.
	 * @param pThresh is the minimum probability for which matches are
	 * generated.
	 * @param model is the syntactic model used for probability-based matching.
	 */
	public DictionaryNameFinder(OntologyDatabase db, String documentText,
			double pThresh, SyntacticProbabilityModel model) {
		this.queryBuffer = new HashMap();
		this.db = db;
		this.documentText = documentText;
		this.pThresh = pThresh;
		this.model = model;
	}

	/**
	 * Automatically processes all words that occur in the documentText given in
	 * the constructor. This function is currently only used for testing
	 * purposes.
	 */
	public void processDocumentText() {
		Set<WordPosition> words = DictionaryAlgorithms.getWords(documentText);
		for (WordPosition word : words) {
			processWord(word.getWord(), word.getBegin());
		}
	}

	/**
	 * Searches for the given word in the dictionary (batched).
	 *
	 * @param word is the word that should be processed.
	 * @param offset is the position of the first character in the text.
	 */
	public void processWord(String word, int offset) {
		// Insert the given word into the query buffer
		Collection<WordPosition> positions = queryBuffer.get(word);
		if (positions == null) {
			positions = new LinkedList();
			queryBuffer.put(word, positions);
		}
		positions.add(new WordPosition(documentText, offset, offset + word.length()));

		// Flush the buffer if the minimum word count is reached
		if (queryBuffer.size() >= MIN_WORD_COUNT) {
			flushQueryBuffer();
		}
	}

	/**
	 * Calculates all matches for the given words.
	 *
	 * @return the matches for all words that have been processed.
	 */
	public Collection<DictionaryMatch> calculateMatches() {
		// Flush all remaining words in the query buffer
		flushQueryBuffer();

		// Create the result list
		Collection<DictionaryMatch> res = new ArrayList();

		// Iterate over each word group and fetch the possible dictionary name
		// matches, add them to the result.
		for (WordGroup wordGroup : wordGroupMap.getWordGroups()) {
			res.addAll(wordGroup.deriveReferences(documentText, pThresh, model));
		}

		return res;
	}
}
