/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.dictionary_name_finder;

import de.unibi.techfak.scie.ontology.DictionaryMatch;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;

/**
 * The OntologyWordGroupMap class is used to organize the dictionary matches that
 * are present in the text.
 */
class WordGroupMap {

	/**
	 * Actual map used to organize the entries.
	 */
	private final Map<String, Map<Long, Map<String, WordGroup>>> map;

	/**
	 * Collection that contains all word groups.
	 */
	private final Collection<WordGroup> wordGroups;

	/**
	 * Constructor of the WordGroupMap class.
	 */
	public WordGroupMap() {
		this.map = new HashMap();
		this.wordGroups = new LinkedList();
	}

	/**
	 * Inserts a word into the map.
	 *
	 * @param word is the word reference that should be inserted into the map.
	 */
	public void insertWord(WordReference wordRef) {
		// Fetch all three discriminating properties of the word reference
		final DictionaryMatch match = wordRef.getMatch();
		final String ontologyName = match.getNode().getOntology().getName();
		final long nodeId = match.getNode().getId();
		final String nodeName = match.getNameDescr().getName();

		// Insert it into the correct ontology
		Map<Long, Map<String, WordGroup>> ontology = map.get(ontologyName);
		if (ontology == null) {
			ontology = new HashMap();
			map.put(ontologyName, ontology);
		}

		// Check whether the node id already exists
		Map<String, WordGroup> node = ontology.get(nodeId);
		if (node == null) {
			node = new TreeMap();
			ontology.put(nodeId, node);
		}

		// Check whether the name already exists, if not insert it
		WordGroup group = node.get(nodeName);
		if (group == null) {
			group = new WordGroup(match.getNameDescr(), match.getNode());
			node.put(nodeName, group);

			// Add the group to the outer word group list
			this.wordGroups.add(group);
		}
		group.addWordReference(wordRef);
	}

	/**
	 * Returns a reference to the word groups.
	 *
	 * @return the word groups that were generated while inserting word
	 * references into the WordGroupMap.
	 */
	public Collection<WordGroup> getWordGroups() {
		return wordGroups;
	}

	/**
	 * Clears all internal maps and the word group list.
	 */
	public void clear() {
		wordGroups.clear();
		map.clear();
	}

};
