/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import java.util.Objects;

/**
 * The Association class represents a directed association between two nodes in
 * an ontology.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public final class Association {

	/**
	 * Node from which the association originates.
	 */
	private final NodeHandle source;

	/**
	 * Node at which the association points.
	 */
	private final NodeHandle target;

	/**
	 * Type of the association.
	 */
	private final AssociationType type;

	/**
	 * Constructor of the Association class. Note that both given nodes have to
	 * be in the same ontology! Links between ontologies are not possible.
	 *
	 * @param source is the node from which the association originates.
	 * @param target is the node at which the association points.
	 * @param type describes the type of the association.
	 */
	public Association(NodeHandle source, NodeHandle target,
			AssociationType type) {
		if (!Objects.equals(source.getOntology(), target.getOntology())) {
			throw new InvalidAssociationException(
					"Given nodes belong to different ontologies!");
		}
		this.source = source;
		this.target = target;
		this.type = type;
	}

	/**
	 * Constructor of the Association class. Note that both given nodes have to
	 * be in the same ontology! Links between ontologies are not possible.
	 *
	 * @param source is the node from which the association originates.
	 * @param target is the node at which the association points.
	 * @param typeValue describes the type of the association as the value that
	 * corresponds to each AssociationType enumeration item.
	 */
	public Association(NodeHandle source, NodeHandle target,
			int typeValue) {
		this(source, target, AssociationType.getType(typeValue));
	}

	/**
	 * Constructor of the Association class. The type of the Association is
	 * chosen as AssociationType.DEFAULT. Note that both given nodes have to be
	 * in the same ontology! Links between ontologies are not possible.
	 *
	 * @param source is the node from which the association originates.
	 * @param target is the node at which the association points.
	 */
	public Association(NodeHandle source, NodeHandle target) {
		this(source, target, AssociationType.DEFAULT);
	}

	/**
	 * Constructor of the Association class.
	 *
	 * @param ontology is the ontology the association belongs to.
	 * @param sourceId is the id of the source node in the given ontology.
	 * @param targetId is the id of the target node in the given ontology.
	 * @param type is the type of the association.
	 */
	public Association(OntologyHandle ontology, int sourceId, int targetId,
			AssociationType type) {
		this(new NodeHandle(ontology, sourceId), new NodeHandle(ontology,
				targetId), type);
	}

	/**
	 * Constructor of the Association class. The type of the association is
	 * chosen as AssociationType.DEFAULT
	 *
	 * @param ontology is the ontology the association belongs to.
	 * @param sourceId is the id of the source node in the given ontology.
	 * @param targetId is the id of the target node in the given ontology.
	 */
	public Association(OntologyHandle ontology, int sourceId, int targetId) {
		this(new NodeHandle(ontology, sourceId), new NodeHandle(ontology,
				targetId));
	}

	/**
	 * Returns a reference to the source node.
	 *
	 * @return a reference to the source node.
	 */
	public NodeHandle getSource() {
		return source;
	}

	/**
	 * Returns a reference to the target node.
	 *
	 * @return a reference to the target node.
	 */
	public NodeHandle getTarget() {
		return target;
	}

	/**
	 * Returns the type of the association.
	 *
	 * @return the type of the association.
	 */
	public AssociationType getType() {
		return type;
	}

	/**
	 * Checks whether the two given objects are equal. Two Associations equal
	 * each other if source, target node are the same.
	 *
	 * @param o is the object for which the equality should be checked.
	 * @return true if the two objects are the same, false otherwise.
	 */
	@Override
	public boolean equals(Object o) {
		if ((o == null) || (o.getClass() != getClass())) {
			return false;
		}

		Association assoc = (Association) o;
		return assoc.source.equals(source) && assoc.target.equals(target);
	}

	/**
	 * Calculates a hash code for the Association from the source node and
	 * target node.
	 *
	 * @return the calculated hash code.
	 */
	@Override
	public int hashCode() {
		return 5 + source.hashCode() * 251 + target.hashCode() * 419;
	}

}
