/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * The NodeDescriptor represents the properties of a single node in the ontology
 * database. Each node can be marked as "root" node (in which case it will be
 * easily accessible using the "getRootNodes" function) and each node can have
 * an arbitrary number of names (instances of NodeName) attached to it.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class NodeDescriptor {

	/**
	 * The handle of the node the descriptor describes.
	 */
	protected final NodeHandle handle;

	/**
	 * The list of names attached to the node.
	 */
	protected final List<NodeName> names;

	/**
	 * Specifies whether this node is a root node or not. Root nodes are easily
	 * accessible using the "getRootNodes" function.
	 */
	protected boolean root;

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(NodeHandle handle, Collection<NodeName> names,
			boolean root) {
		this.handle = handle;
		if (names != null) {
			this.names = new ArrayList(names);
		} else {
			this.names = new ArrayList();
		}
		this.root = root;
	}

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(NodeHandle handle, NodeName name, boolean root) {
		this(handle, Arrays.asList(new NodeName[]{name}), root);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 */
	public NodeDescriptor(NodeHandle handle, Collection<NodeName> names) {
		this(handle, names, false);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 */
	public NodeDescriptor(NodeHandle handle, NodeName name) {
		this(handle, name, false);
	}

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single name that should be attached to the node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(NodeHandle handle, String name, boolean root) {
		this(handle, new NodeName(name), root);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the node the descriptor describes.
	 * @param name is a single name that should be attached to the node.
	 */
	public NodeDescriptor(NodeHandle handle, String name) {
		this(handle, new NodeName(name), false);
	}

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, Collection<NodeName> names,
			boolean root) {
		this(new NodeHandle(handle, id), names, root);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param names is a list of NodeName instances that should be attached to
	 * the node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, Collection<NodeName> names) {
		this(handle, id, names, false);
	}

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, NodeName name,
			boolean root) {
		this(new NodeHandle(handle, id), name, root);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single NodeName instance that should be attached to the
	 * node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, NodeName name) {
		this(handle, id, name, false);
	}

	/**
	 * Constructor of the NodeDescriptor class.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single name that should be attached to the node.
	 * @param root specifies whether the node is a root node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, String name,
			boolean root) {
		this(handle, id, new NodeName(name), root);
	}

	/**
	 * Constructor of the NodeDescriptor class. The node is not a root node.
	 *
	 * @param handle is the handle of the ontology the node should belong to
	 * @param id is the id of the described node.
	 * @param name is a single name that should be attached to the node.
	 */
	public NodeDescriptor(OntologyHandle handle, int id, String name) {
		this(handle, id, new NodeName(name), false);
	}

	/**
	 * Returns the handle of the node this instance describes.
	 *
	 * @return the handle of the node this instance describes.
	 */
	public NodeHandle getHandle() {
		return handle;
	}

	/**
	 * Returns the ontology the described node belongs to.
	 *
	 * @return the ontology the described node belongs to.
	 */
	public OntologyHandle getOntology() {
		return handle.getOntology();
	}

	/**
	 * Returns the id of the described node.
	 *
	 * @return the id of the described node.
	 */
	public int getId() {
		return handle.getId();
	}

	/**
	 * Returns the names of the described node.
	 *
	 * @return the names of the described node.
	 */
	public Collection<NodeName> getNames() {
		return names;
	}

	/**
	 * Returns the NodeName instance of the primary node name. The primary node
	 * name is the node name which is marked as "primary". If no such node name
	 * exists, the first node name from the NodeName list is returned.
	 *
	 * @return the NodeName instance of the primary node name, or null if no
	 * NodeName instance is attached to the descriptor.
	 */
	public NodeName getPrimaryNameDescr() {
		// Check whether there is a name marked as "primary" within the name
		// array.
		NodeName res = null;
		for (NodeName name : names) {
			if (name.isPrimary()) {
				res = name;
				break;
			}
		}

		// If no name was found, simply select the first one in the list.
		if (res == null && !names.isEmpty()) {
			res = names.get(0);
		}

		return res;
	}

	/**
	 * Returns the name part of the primary node name. The primary node name is
	 * the node name which is marked as "primary". If no such node name exists,
	 * the first node name from the NodeName list is returned.
	 *
	 * @return the name part of the primary node name or an empty string if no
	 * names are attached to the node.
	 */
	public String getPrimaryName() {
		NodeName primaryName = getPrimaryNameDescr();
		return primaryName == null ? "" : primaryName.getName();
	}

	/**
	 * Returns whether the node is a root node or not.
	 *
	 * @return true if the node is a root node, false if it is not a root node.
	 */
	public boolean isRoot() {
		return root;
	}

}
