/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver.implementation;

import de.unibi.techfak.scie.ontology.driver.Association;
import de.unibi.techfak.scie.ontology.driver.AssociationDirection;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriverException;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The JDBCDatabaseDriver builds the base class for ontology databases that are
 * based on an relational SQL database.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class JDBCDatabaseDriver extends OntologyDatabaseDriver {

	/**
	 * The JDBCOntologyHandle class is the OntologyHandle type used by the JDBC
	 * database. Despite having a name it also contains a mapping to the
	 * internally used id.
	 */
	protected class JDBCOntologyHandle extends OntologyHandle {

		/**
		 * ID is the internally used id.
		 */
		private final int id;

		/**
		 * Constructor of the JDBCOntologyHandle class.
		 *
		 * @param name is the name of the ontology.
		 * @param id is the internal id of the ontology in the database.
		 */
		public JDBCOntologyHandle(String name, int id) {
			super(name);
			this.id = id;
		}

		/**
		 * Returns the database id.
		 *
		 * @return the database id.
		 */
		public int getId() {
			return id;
		}

	}

	/**
	 * Assembles an exception message for the given SQLException. This method is
	 * neccessary, as the SQLException instances may be chained when batch
	 * queries are used.
	 *
	 * @param sqlException is the exception for which the message should be
	 * generated.
	 * @return the message.
	 */
	private static String getSQLExceptionMessage(SQLException sqlException) {
		final StringBuilder msg = new StringBuilder("SQLException:\n");
		int i = 1;
		while (sqlException != null) {
			msg.append("<");
			msg.append(i++);
			msg.append("> ");
			msg.append(sqlException.getMessage());
			msg.append("\n");
			sqlException = sqlException.getNextException();
		}
		return msg.toString();
	}

	/**
	 * JDBCDatabaseDriverException is a thin wrapper around
	 * DatabaseDriverException that allows for a fast conversion from
	 * SQLException to DatabaseDriverException.
	 */
	protected class JDBCDatabaseDriverException extends OntologyDatabaseDriverException {

		/**
		 * A reference to the actual exception that was thrown.
		 */
		private final Exception parentException;

		/**
		 * Creates an instance of the JDBCDatabaseDriverException class.
		 *
		 * @param msg a prose description of the exception that occured.
		 * @param parentException an (optional) exception that originall
		 * occured.
		 */
		public JDBCDatabaseDriverException(String msg, Exception parentException) {
			super(msg);
			this.parentException = parentException;
		}

		/**
		 * Creates an instance of the JDBCDatabaseDriverException class.
		 *
		 * @param msg a prose description of the exception that occured.
		 */
		public JDBCDatabaseDriverException(String msg) {
			this(msg, null);
		}

		/**
		 * Creates an instance of the JDBCDatabaseDriverException class.
		 *
		 * @param sqlException an instance of SQLException that should be eaten
		 * by this class.
		 */
		public JDBCDatabaseDriverException(SQLException sqlException) {
			this(getSQLExceptionMessage(sqlException), null);
		}

		/**
		 * Returns the parent exception.
		 *
		 * @return the parent exception. May also be null.
		 */
		public Exception getParentException() {
			return parentException;
		}

	}

	/**
	 * Contains a reference to the JDBC database connection handle.
	 */
	private final Connection connection;

	/**
	 * Contains a reference to the object that returns the SQL statements.
	 */
	private final SQLStatements sql;

	/**
	 * Constructor of the JDBCDatabaseDriver class.
	 *
	 * @param connection is a reference to a JDBC database connection.
	 * @param statements is a reference to an instance of the SQLStatements
	 * class which is responsible for generating the SQL statements.
	 */
	public JDBCDatabaseDriver(Connection connection, SQLStatements statements) {
		this.connection = connection;
		this.sql = statements;
	}

	/**
	 * Internal function which is used to fetch an OntologyHandle for the given
	 * ontology name.
	 *
	 * @param name is the name of the ontology that should be fetched.
	 * @return an ontology handle or null if the ontology does not exist.
	 * @throws SQLException
	 */
	private OntologyHandle fetchOntology(String name) throws SQLException {
		try (ResultSet rs = sql.getStatementGetOntologyByName(name).executeQuery()) {
			if (rs.next()) {
				final int id = rs.getInt(1);
				return new JDBCOntologyHandle(name, id);
			}
		}
		return null;
	}

	/**
	 * Internal function which is used to fetch an OntologyHandle for the given
	 * ontology id.
	 *
	 * @param oid is the id of the ontology.
	 * @return an ontology handle or null if the ontology does not exist.
	 * @throws SQLException
	 */
	private OntologyHandle fetchOntologyById(int oid) throws SQLException {
		try (ResultSet rs = sql.getStatementGetOntologyByID(oid).executeQuery()) {
			if (rs.next()) {
				final String name = rs.getString(1);
				return new JDBCOntologyHandle(name, oid);
			}
		}
		return null;
	}

	/**
	 * Checks whether an table with the given name exist. The comparison is case
	 * insensitive.
	 *
	 * @param table is the name of the table.
	 * @return true if the table exists, false otherwise.
	 * @throws SQLException
	 */
	protected boolean tableExists(String table) throws SQLException {
		final DatabaseMetaData meta = connection.getMetaData();
		final String t = table.toLowerCase();
		try (ResultSet res = meta.getTables(null, null, null,
				new String[]{"TABLE"})) {
			while (res.next()) {

				if (res.getString("TABLE_NAME").toLowerCase().equals(t)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Creates all tables that do not exist yet. Not that the schema of existing
	 * tables is not updated!
	 *
	 * @throws SQLException
	 */
	public void initializeSchema() throws SQLException {
		// Iterate over all tables
		for (String table : sql.getTables()) {
			if (!tableExists(table)) {
				String[] cmds = sql.getSchemaSQLForTable(table);
				for (String cmd : cmds) {
					try (Statement stmt = connection.createStatement()) {
						stmt.executeUpdate(cmd);
					}
				}
			}
		}
	}

	/**
	 * Clears the current ontology database. Drops all ontology db tables and
	 * recreates them.
	 */
	@Override
	public void clear() {
		try {
			for (String table : sql.getTables()) {
				try (Statement stmt = connection.createStatement()) {
					stmt.executeUpdate(sql.getSQLDropTableIfExists(table));
				}
			}
			initializeSchema();
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public Collection<OntologyHandle> getOntologies() {
		final Collection<OntologyHandle> res = new ArrayList();
		try {
			try (ResultSet rs = sql.getStatementGetOntologies().executeQuery()) {
				while (rs.next()) {
					final String name = rs.getString(1);
					final int id = rs.getInt(2);
					res.add(new JDBCOntologyHandle(name, id));
				}
			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
		return res;
	}

	@Override
	public OntologyHandle getOntology(String name, boolean create) {
		try {
			OntologyHandle res = fetchOntology(name);
			if (res == null && create) {
				sql.getStatementInsertOntology(name).executeUpdate();
				res = fetchOntology(name);
			}
			return res;
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public Date getOntologyModificationDate(OntologyHandle ontology) {
		final int oid = ((JDBCOntologyHandle) ontology).getId();
		try {
			try (ResultSet rs = sql.getStatementGetOntologyDate(oid).executeQuery()) {
				if (rs.next()) {
					return rs.getTimestamp(1);
				}
			}
			return null;
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void setOntologyModificationDate(OntologyHandle ontology, Date date) {
		final int oid = ((JDBCOntologyHandle) ontology).getId();
		try {
			sql.getStatementSetOntologyDate(oid,
					new java.sql.Timestamp(date.getTime())).executeUpdate();
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void deleteOntology(OntologyHandle ontology) {
		final int oid = ((JDBCOntologyHandle) ontology).getId();
		try {
			connection.setAutoCommit(false);

			sql.getStatementDeleteOntology(oid).executeUpdate();
			sql.getStatementDeleteAssociationsForOntology(oid).executeUpdate();
			sql.getStatementDeleteNodeNamesForOntology(oid).executeUpdate();
			sql.getStatementDeleteDictionaryEntriesForOntology(oid).executeUpdate();
			sql.getStatementDeleteNodesForOntology(oid).executeUpdate();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public Collection<NodeDescriptor> getRootNodes(OntologyHandle ontology) {
		// Prepare the result and fetch the ontology id
		final int oid = ((JDBCOntologyHandle) ontology).getId();

		try {
			// Make sure the ontology exists, if not return null
			if (fetchOntologyById(oid) == null) {
				return null;
			}

			// Fetch the handles of all root nodes and return the corresponding
			// descriptors in a second query.
			try (ResultSet rs = sql.getStatementGetRootNodes(oid).executeQuery()) {
				final List<NodeHandle> handles = new ArrayList();
				while (rs.next()) {
					final int id = rs.getInt(1);
					handles.add(new NodeHandle(ontology, id));
				}
				return getNodes(handles);
			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public List<NodeDescriptor> getNodes(Collection<NodeHandle> handles) {
		List<NodeDescriptor> res = new ArrayList();
		try {
			for (NodeHandle handle : handles) {
				final int oid = ((JDBCOntologyHandle) handle.getOntology()).getId();
				final int id = handle.getId();

				NodeDescriptor node = null;
				try (ResultSet rs = sql.getStatementGetIsRootNode(oid, id).executeQuery()) {
					if (rs.next()) {
						final boolean root = rs.getBoolean(1);
						node = new NodeDescriptor(handle, (List<NodeName>) null, root);
					}
				}

				if (node != null) {
					try (ResultSet rs = sql.getStatementGetNodeNames(oid, id).executeQuery()) {
						while (rs.next()) {
							final String name = rs.getString(1);
							final String type = rs.getString(2);
							final ComparisonMode mode = ComparisonMode.getMode(rs.getInt(3));
							final boolean primary = rs.getBoolean(4);
							node.getNames().add(new NodeName(name,
									type.isEmpty() ? null : type, mode, primary));
						}
					}
				}
				res.add(node);
			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
		return res;
	}

	@Override
	public List<Boolean> queryNodesExist(Collection<NodeHandle> handles) {
		List<Boolean> res = new ArrayList();
		try {
			for (NodeHandle handle : handles) {
				final int oid = ((JDBCOntologyHandle) handle.getOntology()).getId();
				final int id = handle.getId();
				try (ResultSet rs = sql.getStatementGetNodeExists(oid, id).executeQuery()) {
					res.add(rs.next() && rs.getInt(1) > 0);
				}
			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
		return res;
	}

	@Override
	public void storeNodes(Collection<NodeDescriptor> descriptors) {
		try {
			connection.setAutoCommit(false);

			// Delete all given nodes first in order to make sure they no longer
			// exist, then insert them again
			for (NodeDescriptor descr : descriptors) {
				// Fetch ontology id and node id
				final int oid = ((JDBCOntologyHandle) descr.getOntology()).getId();
				final int id = descr.getId();

				// Delete the old node descriptor
				sql.getStatementDeleteNodeID(oid, id).addBatch();
				sql.getStatementDeleteNodeNames(oid, id).addBatch();

				// Write the new one
				sql.getStatementInsertNodeID(oid, id, descr.isRoot()).addBatch();
				for (NodeName name : descr.getNames()) {
					final String n = name.getName();
					final String t = name.getType() == null ? "" : name.getType();
					final int m = name.getComparisonMode().getValue();
					final boolean p = name.isPrimary();
					sql.getStatementInsertNodeName(oid, id, n, t, m, p).addBatch();
				}
			}

			// Execute the batch statements
			sql.getStatement(SQLStatements.SID.deleteNodeID).executeBatch();
			sql.getStatement(SQLStatements.SID.deleteNodeNames).executeBatch();
			sql.getStatement(SQLStatements.SID.insertNodeID).executeBatch();
			sql.getStatement(SQLStatements.SID.insertNodeName).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void deleteNodes(Collection<NodeHandle> nodes) {
		try {
			connection.setAutoCommit(false);

			// Delete all given nodes first in order to make sure they no longer
			// exist, then insert them again
			for (NodeHandle node : nodes) {
				// Fetch ontology id and node id
				final int oid = ((JDBCOntologyHandle) node.getOntology()).getId();
				final int id = node.getId();

				// Prepare the deletion
				sql.getStatementDeleteNodeID(oid, id).addBatch();
				sql.getStatementDeleteNodeNames(oid, id).addBatch();
				sql.getStatementDeleteNodeAssociations(oid, id).addBatch();
				sql.getStatementDeleteDictionaryEntry(oid, id).addBatch();
			}

			// Execute the batch
			sql.getStatement(SQLStatements.SID.deleteNodeID).executeBatch();
			sql.getStatement(SQLStatements.SID.deleteNodeNames).executeBatch();
			sql.getStatement(SQLStatements.SID.deleteNodeAssociations).executeBatch();
			sql.getStatement(SQLStatements.SID.deleteDictionaryEntry).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public List<Collection<Association>> getAssociations(
			Collection<NodeHandle> nodes, int associationTypes,
			AssociationDirection dir) {
		List<Collection<Association>> res = new ArrayList();
		try {
			for (NodeHandle handle : nodes) {
				// Initialize all needed variables
				final Collection<Association> associations = new ArrayList();
				final JDBCOntologyHandle ontology = (JDBCOntologyHandle) handle.getOntology();
				final int oid = ontology.getId();
				final int id = handle.getId();

				// Select the statment depending on the association direction
				PreparedStatement stmt;
				switch (dir) {
					case BOTH:
						stmt = sql.getStatementGetUndirectedAssociations(oid, id, associationTypes);
						break;
					case INBOUND:
						stmt = sql.getStatementGetInboundAssociations(oid, id, associationTypes);
						break;
					case OUTBOUND:
						stmt = sql.getStatementGetOutboundAssociations(oid, id, associationTypes);
						break;
					default:
						res.add(associations);
						continue;
				}

				// Execute the statement and add all associations to the result
				try (ResultSet rs = stmt.executeQuery()) {
					while (rs.next()) {
						final int src = rs.getInt(1);
						final int tar = rs.getInt(2);
						final int type = rs.getInt(3);
						associations.add(new Association(
								new NodeHandle(ontology, src),
								new NodeHandle(ontology, tar),
								type));
					}
					res.add(associations);
				}

			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
		return res;
	}

	@Override
	public void storeAssociations(Collection<Association> associations) {
		try {
			connection.setAutoCommit(false);

			// List which collects all handles for efficient node exists query
			List<NodeHandle> handles = new ArrayList();

			for (Association association : associations) {
				// Fetch ontology id and node id
				final NodeHandle srcNode = association.getSource();
				final NodeHandle tarNode = association.getTarget();
				final int oid = ((JDBCOntologyHandle) srcNode.getOntology()).getId();
				final int src = srcNode.getId();
				final int tar = tarNode.getId();
				final int type = association.getType().getValue();

				// Delete any old association, then insert the new one
				sql.getStatementDeleteAssociation(oid, src, tar).addBatch();
				sql.getStatementInsertAssociation(oid, src, tar, type).addBatch();

				// Store the handles
				handles.add(srcNode);
				handles.add(tarNode);
			}

			// Check whether the handles exist
			boolean mustInsertNodes = false;
			List<Boolean> handlesExist = queryNodesExist(handles);
			for (int i = 0; i < handlesExist.size(); i++) {
				if (!handlesExist.get(i)) {
					// Fetch the node handle and its oid/id
					final NodeHandle node = handles.get(i);
					final int oid = ((JDBCOntologyHandle) node.getOntology()).getId();
					final int id = node.getId();

					// Create the node by inserting it
					sql.getStatementInsertNodeID(oid, id, false).addBatch();
					mustInsertNodes = true;
				}
			}

			// Execute the batch statements
			if (mustInsertNodes) {
				sql.getStatement(SQLStatements.SID.insertNodeID).executeBatch();
			}
			sql.getStatement(SQLStatements.SID.deleteAssociation).executeBatch();
			sql.getStatement(SQLStatements.SID.insertAssociation).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void deleteAssociations(Collection<Association> associations) {
		try {
			connection.setAutoCommit(false);

			for (Association association : associations) {
				// Fetch ontology id and node id
				final NodeHandle srcNode = association.getSource();
				final NodeHandle tarNode = association.getTarget();
				final int oid = ((JDBCOntologyHandle) srcNode.getOntology()).getId();
				final int src = srcNode.getId();
				final int tar = tarNode.getId();

				// Queue the delete statement
				sql.getStatementDeleteAssociation(oid, src, tar).addBatch();
			}

			// Execute the batch
			sql.getStatement(SQLStatements.SID.deleteAssociation).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public Map<String, Collection<NodeDescriptor>> queryDictionary(Set<String> keys) {
		Map<Integer, OntologyHandle> ontologyCache = new HashMap();
		Map<String, Collection<NodeDescriptor>> res = new HashMap();
		try {
			for (String key : keys) {
				// Initialize all needed variables
				final Collection<NodeHandle> handles = new ArrayList();

				// Execute the statement and add all associations to the result
				try (ResultSet rs = sql.getStatementQueryDictionary(key).executeQuery()) {
					while (rs.next()) {
						final int oid = rs.getInt(1);
						final int id = rs.getInt(2);

						// Fetch the ontology handle that belongs to the given
						// id
						if (!ontologyCache.containsKey(oid)) {
							OntologyHandle ontology = fetchOntologyById(oid);
							ontologyCache.put(oid, ontology);
						}
						OntologyHandle ontology = ontologyCache.get(oid);
						if (ontology != null) {
							handles.add(new NodeHandle(ontology, id));
						}
					}
				}
				res.put(key, getNodes(handles));
			}
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
		return res;
	}

	@Override
	public void storeDictionaryEntries(Map<NodeHandle, Set<String>> keys) {
		try {
			connection.setAutoCommit(false);

			for (NodeHandle node : keys.keySet()) {
				// Fetch ontology id and node id
				final int oid = ((JDBCOntologyHandle) node.getOntology()).getId();
				final int id = node.getId();

				// Queue each key insertion
				for (String key : keys.get(node)) {
					sql.getStatementInsertDictionaryEntry(oid, id, key).addBatch();
				}
			}

			// Execute the batch insertion
			sql.getStatement(SQLStatements.SID.insertDictionaryEntry).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void deleteDictionaryEntries(Collection<NodeHandle> nodes) {
		try {
			connection.setAutoCommit(false);

			for (NodeHandle node : nodes) {
				// Fetch ontology id and node id
				final int oid = ((JDBCOntologyHandle) node.getOntology()).getId();
				final int id = node.getId();

				// Add the deletion to the queue
				sql.getStatementDeleteDictionaryEntry(oid, id).addBatch();
			}

			// Execute the batch deletion
			sql.getStatement(SQLStatements.SID.deleteDictionaryEntry).executeBatch();

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException ex) {
			throw new JDBCDatabaseDriverException(ex);
		}
	}

	@Override
	public void close() {
		try {
			connection.close();
		} catch (SQLException ex) {
			// Do nothing here -- we want to close the connection!
		}
	}

	@Override
	public String getSchemaVersion() {
		return "2014-05-21.01";
	}

}
