/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer.mesh;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.importer.AbstractImporter;

/**
 *
 *
 */
public class MeshImporter2 extends AbstractImporter {
	//Tag for a new entry
	private static String ENTRY="SupplementalRecord";
	// Tag for Id of an entry
	private static String ID="SupplementalRecordUI";
	//Tag for Name of an entry
	private static String NAME="SupplementalRecordName";
	//Tag for chemical formular 
	private static String CHEMICAL="CASN1Name";
	//Tag for list of pharmacological actions
	private static String PHARMA="PharmacologicalActionList";
	//Tag for Strings
	private static String STRING="String";
	//Tag for list of synonyms
	private static String SYNONYM="TermList";
	//Tag for qualifier
	private static String QUAL="QualifierUI";
	//Text for qualifier administration & dosage => is a drug
	private static String DRUG="Q000008";
	
	private enum IdKind{
		undefined,
		entryId,
		pharmaRefId;
	}
	private enum StringKind{
		undefined,
		entryName,
		synonymName,
		refName;
	}
	/**
	 * Instanciates the importer for the given database. Uses "drug" as ontology
	 * name.
	 *
	 * @param db is a reference to the database in which the ontology should be
	 * stored.
	 */
	public MeshImporter2(OntologyDatabase db) {
		super(db, "drug2");
	}

	/**
	 * Flushes all data from desc xml(Mesh) into db.
	 *
	 * @param onto refers to the Ontology into which the datbase should be
	 * imported.
	 * @param parser the XMLStreamReader for the desc xml
	 * @throws XMLStreamException
	 */
	@Override
	protected void parseXML(OntologyHandle onto, XMLStreamReader parser) throws XMLStreamException {
		IdKind idKind=IdKind.undefined;
		StringKind stringKind=StringKind.undefined;
		boolean isDrug=false;
		boolean isName=false;
		int taxonId=-1;
		int refID=-1;
		int imported = 0;
		int parsed=0;
		List<NodeName> taxonNames = new LinkedList();
		HashMap<Integer, String> references=new HashMap<Integer,String>();
		while (parser.hasNext()) {
			switch (parser.getEventType()) {

				case XMLStreamConstants.END_DOCUMENT:
					parser.close();
					break;

				case XMLStreamConstants.START_ELEMENT:
					if(parser.getLocalName().equals(ENTRY)){
						idKind=IdKind.entryId;
						isName=true;
					}else if(parser.getLocalName().equals(PHARMA)){
						idKind=IdKind.pharmaRefId;
						stringKind=StringKind.refName;
						isDrug=true;
					}else if(parser.getLocalName().equals(ID)){
						switch (idKind) {
						case entryId:
							//set Id
							String text=parser.getElementText().trim();
							taxonId=Integer.parseInt(text.substring(1));
							idKind=IdKind.undefined;
							break;
						case pharmaRefId:
							//add ref
							refID=Integer.parseInt(parser.getElementText().trim().substring(1));
							break;
						default:
							break;
						}
					}else if(parser.getLocalName().equals(NAME)){
						if(isName){
							stringKind=StringKind.entryName;
						}
					}else if(parser.getLocalName().equals(SYNONYM)){
						stringKind=StringKind.synonymName;
					}else if(parser.getLocalName().equals(STRING)){
						switch(stringKind){
							case entryName:
								//set Name
								taxonNames.add(new NodeName(parser.getElementText(), "NAME"));
								break;
							case synonymName:
								//add Synonym
								taxonNames.add(new NodeName(parser.getElementText(), "SYNONYM"));
								break;
							case refName:
								//add name of reference
								references.put(refID, parser.getElementText());
								refID=-1;
							default:
								break;
						}
					}else if(parser.getLocalName().equals(QUAL)){
						if(parser.getElementText().equals(DRUG)){
							isDrug=true;
						}
					}
					break;
				case XMLStreamConstants.END_ELEMENT:
					if(parser.getLocalName().equals(ENTRY)){
						idKind=IdKind.undefined;
						if(isDrug){
							if (taxonId >= 0 && taxonNames.size() > 0) {
								Node node = new Node(db, onto, taxonId, taxonNames);
								db.storeNode(node);
								for (Map.Entry<Integer, String> entry : references.entrySet()) {
							    Integer key = entry.getKey();
							    String value = entry.getValue();
							    Node ref= new Node(db,onto,key.intValue(),new NodeName(value, "ACTION",ComparisonMode.NONE));
							    db.storeNode(ref);
							    node.addAssociationFrom(ref, AssociationType.IS_A);
								}
							}
							imported++;
						}
						parsed++;
						taxonId = -1;
						taxonNames.clear();
						references.clear();
						isDrug=false;
						if (parsed % 1000 == 0) {
							System.out.println(parsed + " items parsed, "+imported+" Drugs imported");
						}
					}else if(parser.getLocalName().equals(PHARMA)){
						idKind=IdKind.undefined;
						stringKind=StringKind.undefined;
					}else if(parser.getLocalName().equals(NAME)){
						stringKind=StringKind.undefined;
						isName=false;
					}else if(parser.getLocalName().equals(SYNONYM)){
						stringKind=StringKind.undefined;
					}
				default:
					break;
			}
			parser.next();
		}
	}
}
