/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer.small;

import java.util.LinkedList;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.importer.AbstractImporter;

public class SmallImporter extends AbstractImporter {

	public SmallImporter(OntologyDatabase db, String ontologyName) {
		super(db, ontologyName);
	}

	@Override
	protected void parseXML(OntologyHandle onto, XMLStreamReader parser)
			throws XMLStreamException {
		// Current node information
		int taxonId = -1;
		List<NodeName> taxonNames = new LinkedList();
		List<AssociationList> associations = new LinkedList();
		AssociationList currentAssociations = null;
		while (parser.hasNext()) {
			// Fetch the tag name
			final int eventType = parser.getEventType();
			String tagName = "";
			if (eventType == XMLStreamConstants.START_ELEMENT
					|| eventType == XMLStreamConstants.END_ELEMENT) {
				tagName = parser.getLocalName();
			}
			switch (eventType) {
				case XMLStreamConstants.END_DOCUMENT:
					parser.close();
					break;

				case XMLStreamConstants.START_ELEMENT:
					switch (tagName) {
						case "node":
							String id = parser.getAttributeValue(null, "id");
							taxonId = Integer.parseInt(id);
							currentAssociations = new AssociationList(taxonId);
							break;
						case "name":
							String type = parser.getAttributeValue(null, "type");
							if (type == null || type.equals("")) {
								type = null;
							}
							String mode = parser.getAttributeValue(null, "mode");
							if (mode == null || mode.equals("")) {
								mode = "fuzzy";
							}
							String primary = parser.getAttributeValue(null, "primary");
							boolean prim = false;
							if (primary != null && primary.equals("true")) {
								prim = true;
							}
							String name = parser.getElementText();
							ComparisonMode realMode = ComparisonMode.getMode(mode);
							if (realMode == null) {
								throw new RuntimeException("Invalid comparison mode \"" + mode + "\"!");
							}
							taxonNames.add(new NodeName(name, type, realMode, prim));
							break;
						case "association":
							String target = parser.getAttributeValue(null, "to");
							type = parser.getAttributeValue(null, "type");
							if (type == null || type.equals("")) {
								type = "default";
							}
							AssociationType realType = AssociationType.getType(type);
							if (realType == null) {
								throw new RuntimeException("Invalid association type \"" + type + "\"!");
							}
							int targetId = Integer.parseInt(target);
							currentAssociations.add(new AbstractAssociation(targetId, realType));
							break;
						default:
							break;
					}
					break;

				case XMLStreamConstants.ATTRIBUTE:

				case XMLStreamConstants.END_ELEMENT:
					switch (tagName) {
						case "node":
							boolean onePrim = false;
							for (NodeName name : taxonNames) {
								if (name.isPrimary()) {
									if (onePrim) {
										throw new RuntimeException("Only one Primary-Name allowed");
									}
									onePrim = true;
								}
							}
							Node node = new Node(db, onto, taxonId, taxonNames);
							db.storeNode(node);
							System.out.println("imported: " + node.getPrimaryName() + " #" + taxonId);
							taxonId = -1;
							taxonNames.clear();
							associations.add(currentAssociations);
							break;
						default:
							break;
					}
					break;
				default:
					break;
			}
			parser.next();
		}
		for (AssociationList asso : associations) {
			asso.flush(db, onto);
		}
	}

}
