/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model;

/**
 * This calculates the syntactic probability based on the char distance.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class CharBasedSyntacticProbabilityModel implements
		SyntacticProbabilityModel {

	private final double sigma;

	/**
	 *
	 * @param sigma Sigma determines after how many characters distance the
	 * syntactic probability is reduced to 63%.
	 */
	public CharBasedSyntacticProbabilityModel(double sigma) {
		this.sigma = sigma;
	}

	/**
	 * Calculates not the actual normal distribution N(input|0,sigma) /
	 * N(0|0,sigma)
	 *
	 * @param input the input for the density function.
	 * @return the value of N(input|0,sigma) / N(0|0,sigma)
	 */
	public double calculateNormalDistribution(final double input) {
		return Math.exp(-Math.pow(input / sigma, 2) / 2);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double calculateSyntacticProbability(AnnotationCandidate a,
			AnnotationCandidate b) {
		final int distance;
		final int xBegin = a.getBegin();
		final int xEnd = a.getEnd();
		final int yBegin = b.getBegin();
		final int yEnd = b.getEnd();
		if (xBegin < yBegin) {
			if (xEnd <= yBegin) {
				//if x is before y, we define the distance as yBegin - xEnd
				distance = yBegin - xEnd;
			} else {
				//if both overlap we take the minimum of the distance between both beginnings and both ends.
				distance = Math.min(Math.abs(xBegin - yBegin), Math.abs(xEnd -
						yEnd));
			}
		} else {
			if (yEnd <= xBegin) {
				//if y is before x, we define the distance as xBegin - yEnd
				distance = xBegin - yEnd;
			} else {
				//if both overlap we take the minimum of the distance between both beginnings and both ends.
				distance = Math.min(Math.abs(xBegin - yBegin), Math.abs(xEnd -
						yEnd));
			}
		}

		return calculateNormalDistribution(distance);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getMaximumCharBasedRange() {
		return 3 * (int) sigma;
	}
}
