/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model.uima;

import de.unibi.techfak.scie.probability_model.AnnotationCandidate;
import de.unibi.techfak.scie.probability_model.AnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.CharBasedSyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

/**
 * This class provides some convenience functions to use the AnnotationCompositionUtil with UIMA
 * annotations.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class UIMAAnnotationCompositionUtil {

	public static class UIMAAnnotationCandidate<X extends Annotation> implements
			AnnotationCandidate {

		private final X annotation;

		public UIMAAnnotationCandidate(X annotation) {
			this.annotation = annotation;
		}

		@Override
		public int getBegin() {
			return annotation.getBegin();
		}

		@Override
		public int getEnd() {
			return annotation.getEnd();
		}

		public X getAnnotation() {
			return annotation;
		}
	};

	public static abstract class UIMASemanticProbabilityModel<X extends Annotation>
			implements SemanticProbabilityModel<UIMAAnnotationCandidate<X>> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public double calculateSemanticProbability(
				UIMAAnnotationCandidate<X> candidate) {
			return calculateSemanticProbability(candidate.getAnnotation());
		}

		/**
		 * This should calculate the semantic probabilty of an AnnotationCandidate belonging to
		 * another
		 *
		 * If additional information is required to derive this value set those parameters in some
		 * other way (e.g. the given AnnotationCandidate implementation).
		 *
		 * @param candidate an AnnotationCandidate of Type X.
		 * @return likelihood probabilty of an AnnotationCandidate belonging to a larger composit
		 * annotation
		 */
		public abstract double calculateSemanticProbability(X candidate);
	}

	/**
	 * This calculates the AnnotationCandidate with the maximum probability to belong to some given
	 * core for a composit annotation and returns it with the probability (without normalization!)
	 * attached.
	 *
	 * The probability is defined as the product of the syntactic and the semantic probability of
	 * those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by implementing a
	 * SemanticProbabilityModel for the AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations in the tests.
	 * Details are described in the documentation of the calculateSyntacticProbability-method.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param semanticProbModel a model that is able to calculate the probability of two annotation
	 * candidates belonging together.
	 * @param candidateClass the class of the candidate.
	 * @param jcas the current JCas object
	 * @param syntacticSigma this influences how the syntactic probability is calculated. For an
	 * in-depth documentation have a look at the calculateSyntacticProbability-method.
	 * @return returns the most probable candidate with the probability attached.
	 */
	public static <X extends Annotation> ProbableHypothesis<UIMAAnnotationCandidate<X>> calculateMaxProbHypothesis(
			final Annotation core,
			final UIMASemanticProbabilityModel<X> likelihoodModel,
			final Class<? extends X> candidateClass,
			final JCas jcas, final double syntacticSigma) {

		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateMaxProbHypothesis(core, syntacticProbModel,
				likelihoodModel, candidateClass, jcas);
	}

	/**
	 * This calculates the AnnotationCandidate with the maximum probability to belong to some given
	 * core for a composit annotation and returns it with the probability (without normalization!)
	 * attached.
	 *
	 * The probability is defined as the product of the syntactic and the semantic probability of
	 * those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by implementing a
	 * SemanticProbabilityModel for the AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations in the tests.
	 * Details are described in the documentation of the calculateSyntacticProbability-method.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param syntacticProbModel a model that is able to calculate the probabiliy of two
	 * AnnotationCandidates belonging together from a syntactic perspective.
	 * @param semanticProbModel a model that is able to calculate the probability of two annotation
	 * candidates belonging together from a semantic perspective.
	 * @param candidateClass the class of the candidate.
	 * @param jcas the current JCas object
	 * @return returns the most probable candidate with the probability attached.
	 */
	public static <X extends Annotation> ProbableHypothesis<UIMAAnnotationCandidate<X>> calculateMaxProbHypothesis(
			final Annotation core,
			final SyntacticProbabilityModel syntacticProbModel,
			final UIMASemanticProbabilityModel<X> semanticProbModel,
			final Class<? extends X> candidateClass,
			final JCas jcas) {

		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		//select all candidates in the range of three standard deviances. Everything else is
		//approximated by a probability of zero.
		final int effectiveRange = syntacticProbModel.getMaximumCharBasedRange();
		final Collection<? extends X> candidates = jcasUtil.selectCovered(candidateClass,
				core.getBegin() - effectiveRange, core.getEnd() + effectiveRange);

		return AnnotationCompositionUtil.calculateMaxProbHypothesis(
				new UIMAAnnotationCandidate<>(core), syntacticProbModel,
				semanticProbModel, wrap(candidates));
	}

	/**
	 * For any of the given candidates the probability is calculated, that this candidate belongs to
	 * the given core annotation. The probability is defined as the product of the syntactic and the
	 * semantic probability of those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by implementing a
	 * SemanticProbabilityModel for the AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations in the tests.
	 * Details are described in the documentation of the calculateSyntacticProbability-method.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param semanticProbModel a model that is able to calculate the probability of two annotation
	 * candidates belonging together.
	 * @param candidateClass the class of the candidate.
	 * @param jcas the current JCas object
	 * @param syntacticSigma this influences how the syntactic probability is calculated. For an
	 * in-depth documentation have a look at the calculateSyntacticProbability-method.
	 * @return A subset of the given candidates sorted w.r.t. their probability. The most probable
	 * hypothesis comes first.
	 */
	public static <X extends Annotation> ArrayList<ProbableHypothesis<UIMAAnnotationCandidate<X>>> calculateProbableHypotheses(
			final Annotation core,
			final UIMASemanticProbabilityModel<X> semanticModel,
			final Class<? extends X> candidateClass,
			final JCas jcas, final double syntacticSigma) {
		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateProbableHypotheses(core, syntacticProbModel,
				semanticModel, candidateClass, jcas, syntacticSigma);
	}

	/**
	 * For any of the given candidates the probability is calculated, that this candidate belongs to
	 * the given core annotation. The probability is defined as the product of the syntactic and the
	 * semantic probability of those two annotations belonging together.
	 *
	 * The semantic probability should be given by the user of this function by implementing a
	 * SemanticProbabilityModel for the AnnotationCandidate-Class.
	 *
	 * The syntactic probability is related to the distance of both annotations in the tests.
	 * Details are described in the documentation of the calculateSyntacticProbability-method.
	 *
	 * @param core the core annotation for a bigger Annotation Compositon.
	 * @param syntacticModel a model that is able to calculate the probabiliy of two
	 * AnnotationCandidates belonging together from a syntactic perspective.
	 * @param semanticProbModel a model that is able to calculate the probability of two annotation
	 * candidates belonging together.
	 * @param candidateClass the class of the candidate.
	 * @param jcas the current JCas object
	 * @param syntacticSigma this influences how the syntactic probability is calculated. For an
	 * in-depth documentation have a look at the calculateSyntacticProbability-method.
	 * @return A subset of the given candidates sorted w.r.t. their probability. The most probable
	 * hypothesis comes first.
	 */
	public static <X extends Annotation> ArrayList<ProbableHypothesis<UIMAAnnotationCandidate<X>>> calculateProbableHypotheses(
			final Annotation core,
			final SyntacticProbabilityModel syntacticModel,
			final UIMASemanticProbabilityModel<X> semanticModel,
			final Class<? extends X> candidateClass,
			final JCas jcas, final double syntacticSigma) {

		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		//select all candidates in the range of three standard deviances. Everything else is
		//approximated by a probability of zero.
		final int effectiveRange = syntacticModel.getMaximumCharBasedRange();
		final Collection<? extends X> candidates = jcasUtil.selectCovered(candidateClass,
				core.getBegin() - effectiveRange, core.getEnd() + effectiveRange);
		return AnnotationCompositionUtil.calculateProbableHypotheses(
				new UIMAAnnotationCandidate<>(core), syntacticModel,
				semanticModel, wrap(candidates));
	}

	/**
	 * This calculates all composita that are possible to create given a certain core annotation and
	 * a list of candidates for each slots of the compositum.
	 *
	 * Please note that this method assumes that the compositum is structured as follows:
	 *
	 * <code>class CompositumClass {
	 *
	 *	CoreClass core;
	 *	Slot1Class slot1Instance;
	 *	Slot2Class slot2Instance;
	 *	...
	 * }</code>
	 *
	 * The user is required to provide a SemanticProbabilityModel and a collection of a priori
	 * candidates for each slot.
	 *
	 * Also she is required to construct a class that is able to actually create the compositum from
	 * one candidate for each slot.
	 *
	 * This uses the CharBasedSyntacticProbabilityModel per default.
	 *
	 * @param core the core annotation that shall be the center of the new compositum.
	 * @param semanticModels An array of SemanticProbabilityModels, one for each slot.
	 * @param slotClasses An Array of classes, one for each slot.
	 * @param compositor An AnnotationCompositor instance to construct the resulting composita from
	 * all possible hypothesis combinations.
	 * @param jcas the current JCas instance.
	 * @param syntacticSigma this influences how the syntactic probability is calculated. For an
	 * in-depth documentation have a look at the calculateSyntacticProbability-method.
	 * @return a collection of constructed composita as it is done by the given compositor.
	 */
	public static <Y extends Annotation> List<Y> calculateComposita(
			final Annotation core,
			final UIMASemanticProbabilityModel[] semanticModels,
			final Class[] slotClasses,
			final UIMACompositor<Y> compositor,
			final JCas jcas, final double syntacticSigma) {
		final SyntacticProbabilityModel syntacticProbModel =
				new CharBasedSyntacticProbabilityModel(syntacticSigma);
		return calculateComposita(core, syntacticProbModel, semanticModels,
				slotClasses, compositor, jcas);
	}

	/**
	 * This calculates all composita that are possible to create given a certain core annotation and
	 * a list of candidates for each slots of the compositum.
	 *
	 * Please note that this method assumes that the compositum is structured as follows:
	 *
	 * <code>class CompositumClass {
	 *
	 *	CoreClass core;
	 *	Slot1Class slot1Instance;
	 *	Slot2Class slot2Instance;
	 *	...
	 * }</code>
	 *
	 * The user is required to provide a SemanticProbabilityModel and a collection of a priori
	 * candidates for each slot.
	 *
	 * Also she is required to construct a class that is able to actually create the compositum from
	 * one candidate for each slot.
	 *
	 * @param core the core annotation that shall be the center of the new compositum.
	 * @param syntacticModel a model that is able to calculate the probabiliy of two
	 * AnnotationCandidates belonging together from a syntactic perspective.
	 * @param semanticModels An array of SemanticProbabilityModels, one for each slot.
	 * @param slotClasses An Array of classes, one for each slot.
	 * @param compositor An AnnotationCompositor instance to construct the resulting composita from
	 * all possible hypothesis combinations.
	 * @param jcas the current JCas instance.
	 * @return a collection of constructed composita as it is done by the given compositor.
	 */
	public static <Y extends Annotation> List<Y> calculateComposita(
			final Annotation core,
			final SyntacticProbabilityModel syntacticModel,
			final UIMASemanticProbabilityModel[] semanticModels,
			final Class[] slotClasses,
			final UIMACompositor<Y> compositor,
			final JCas jcas) {
		compositor.setJcas(jcas);
		//check input validity.
		if (semanticModels.length != slotClasses.length) {
			throw new RuntimeException(
					"We need one SemanticProbabilityModel for each class");
		}

		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		//get all possible hypotheses for the single slots.
		final int effectiveRange = syntacticModel.getMaximumCharBasedRange();
		final Collection[] slotHypos = new Collection[slotClasses.length];
		for (int i = 0; i < slotClasses.length; i++) {
			List candidates = jcasUtil.selectCovered(slotClasses[i],
					core.getBegin() - effectiveRange, core.getEnd() + effectiveRange);
			slotHypos[i] = wrap(candidates);
		}
		//if we have all those lists call the AnnotationCompositionUtil.
		return AnnotationCompositionUtil.calculateComposita(
				new UIMAAnnotationCandidate(core), syntacticModel,
				semanticModels, slotHypos, compositor);
	}

	public static <X extends Annotation> ArrayList<UIMAAnnotationCandidate<X>> wrap(
			Collection<? extends X> annotations) {
		final ArrayList<UIMAAnnotationCandidate<X>> resultList =
				new ArrayList<>();
		for (final X annotation : annotations) {
			resultList.add(new UIMAAnnotationCandidate<>(annotation));
		}
		return resultList;
	}
}
