/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model.uima;

import de.unibi.techfak.isy.lppba.descriptors.Paragraph;
import de.unibi.techfak.isy.lppba.descriptors.Sentence;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import de.unibi.techfak.scie.probability_model.AnnotationCandidate;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import java.util.Collection;
import java.util.HashMap;
import java.util.NavigableMap;
import java.util.TreeMap;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

/**
 * This calculates the syntactic probability based on the distance in words and some other syntactic
 * considerations like sentence boundaries.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class UIMAWordSyntacticProbabilityModel implements
		SyntacticProbabilityModel {

	private double sigma;
	private final JCas jcas;
	/**
	 * We punish if both Annotations are not in the same sentence.
	 */
	public int sentenceMalus = 3;
	/**
	 * We punish if both Annotations are not in the same paragraph.
	 */
	public int paragraphMalus = 10;

	public UIMAWordSyntacticProbabilityModel(JCas jcas) {
		this.jcas = jcas;
	}

	/**
	 *
	 * @param sigma Sigma determines after how many <em>words</em> distance the syntactic
	 * probability is reduced to 63%.
	 */
	public UIMAWordSyntacticProbabilityModel(double sigma, JCas jcas) {
		this.sigma = sigma;
		this.jcas = jcas;
	}

	/**
	 *
	 * @param sigma Sigma determines after how many <em>words</em> distance the syntactic
	 * probability is reduced to 63%.
	 */
	public void setSigma(double sigma) {
		this.sigma = sigma;
	}

	/**
	 *
	 * @return Sigma determines after how many <em>words</em> distance the syntactic probability is
	 * reduced to 63%.
	 */
	public double getSigma() {
		return sigma;
	}

	/**
	 *
	 * @return a number specifying how many words are added to the distance if a sentence boundary
	 * is in the way.
	 */
	public int getSentenceMalus() {
		return sentenceMalus;
	}

	/**
	 *
	 * @param sentenceMalus a number specifying how many words are added to the distance if a
	 * sentence boundary is in the way.
	 */
	public void setSentenceMalus(int sentenceMalus) {
		this.sentenceMalus = sentenceMalus;
	}

	/**
	 *
	 * @return a number specifying how many words are added to the distance if a paragraph boundary
	 * is in the way.
	 */
	public int getParagraphMalus() {
		return paragraphMalus;
	}

	/**
	 *
	 * @param paragraphMalus a number specifying how many words are added to the distance if a
	 * paragraph boundary is in the way.
	 */
	public void setParagraphMalus(int paragraphMalus) {
		this.paragraphMalus = paragraphMalus;
	}

	/**
	 * Calculates not the actual normal distribution N(input|0,sigma) / N(0|0,sigma)
	 *
	 * @param input the input for the density function.
	 * @return the value of N(input|0,sigma) / N(0|0,sigma)
	 */
	public double calculateNormalDistribution(final double input) {
		return Math.exp(-Math.pow(input / sigma, 2) / 2);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double calculateSyntacticProbability(AnnotationCandidate a,
			AnnotationCandidate b) {

		final int aBegin = a.getBegin();
		final int aEnd = a.getEnd();
		final int bBegin = b.getBegin();
		final int bEnd = b.getEnd();
		/* 
		 * first treat the special cases that one of the annotations is covering
		 * the other or overlapping with it. If so we return 1.
		 */
		final AnnotationCandidate first;
		final AnnotationCandidate snd;

		if (aBegin < bBegin) {
			if (aEnd <= bBegin) {
				//a is before b
				first = a;
				snd = b;
			} else {
				//if both overlap we return 1
				return 1;
			}
		} else {
			if (bEnd <= aBegin) {
				//b is before a
				first = b;
				snd = a;
			} else {
				//if both overlap we return 1.
				return 1;
			}
		}
		/* 
		 * get the words between both annotations. The number of those words is
		 * the first attempt at a distance.
		 */
		int distance = getDistanceInCoveringClass(Token.class, jcas, first, snd);
		/*
		 * If there are also Sentences in between, we punish that additionally.
		 */
		distance += sentenceMalus * getDistanceInCoveringClass(Sentence.class, jcas, first, snd);

		/*
		 * We do the same for paragraphs.
		 */
		distance += paragraphMalus * getDistanceInCoveringClass(Paragraph.class, jcas, first, snd);

		return calculateNormalDistribution(distance);
	}
	/**
	 * This is a map containing Buffers that each store Annotations of a certain type indiced by
	 * their starting index in the text. This is unfortunately necessary to speed up the syntactic
	 * distance search because UIMA itself is far too slow.
	 */
	private HashMap<Class<? extends Annotation>, TreeMap<Integer, Annotation>> caches =
			new HashMap<>();

	/**
	 * This calculates the distance of two annotations counted in the instance of the given Class.
	 * Please note that this method implicitly assumes that the annotations of the given class are
	 * directly following each other and that only one annotation of that class covers a given
	 * region at a time. This is only garantueed for primitive annotations of the document structure
	 * like Word, Sentence, Paragraph, etc.
	 *
	 * @param <X>
	 * @param annotationClass
	 * @param jcas
	 * @param first
	 * @param snd
	 * @return
	 */
	public <X extends Annotation> int getDistanceInCoveringClass(
			Class<X> annotationClass, JCas jcas, AnnotationCandidate first,
			AnnotationCandidate snd) {
		final int firstBegin = first.getBegin();
		final int sndEnd = snd.getEnd();
		//first check for the correct buffer.
		TreeMap<Integer, Annotation> cache = caches.get(annotationClass);
		if (cache == null) {
			//if we have no buffer yet, get it.
			final Collection<X> annotationInstances = JCasUtil.select(jcas,
					annotationClass);
			cache = new TreeMap<>();
			for (final X anno : annotationInstances) {
				cache.put(anno.getBegin(), anno);
			}
			caches.put(annotationClass, cache);
		}
		//If we have no instances of the given class, we return 0.
		if (cache.isEmpty()) {
			return 0;
		}
		//get the covering annotations.
		final Integer coveringStart;
		if (cache.containsKey(firstBegin)) {
			coveringStart = firstBegin;
		} else {
			coveringStart = cache.lowerKey(firstBegin);
		}
		final Integer coveringEnd;
		if (cache.containsKey(sndEnd)) {
			coveringEnd = sndEnd;
		} else {
			coveringEnd = cache.lowerKey(sndEnd);
		}
		//if we look at a point in the text where there is no instance of the given class yet, we return 0.
		if(coveringStart == null){
			return 0;
		}
		
		final NavigableMap<Integer, Annotation> covering;
		try {
			covering =
					cache.subMap(coveringStart, true, coveringEnd, true);
		} catch (NullPointerException ex) {
			throw new RuntimeException(ex);
		}
		//In the special case that no Annotation was found at all we return 0.
		if (covering.isEmpty()) {
			return 0;
		}

		//Normally we give back the number of those annotations -1.
		return covering.size() - 1;
	}
	/**
	 * We approximate the number of chars per word with 7 (to be somewhat sure) to give a rough
	 * approximation of what region this Model is considering effectively.
	 */
	public static final int APPROXCHARSPERWORD = 7;

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getMaximumCharBasedRange() {
		return 3 * APPROXCHARSPERWORD * (int) sigma;
	}
}
