/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.structure;

import de.unibi.techfak.isy.lppba.descriptors.DocumentPage;
import de.unibi.techfak.isy.lppba.descriptors.Paragraph;
import de.unibi.techfak.isy.lppba.descriptors.Subscript;
import de.unibi.techfak.isy.lppba.descriptors.Superscript;
import de.unibi.techfak.isy.lppba.descriptors.TextBlock;
import de.unibi.techfak.scie.document_import.DocumentImporter;
import de.unibi.techfak.scie.document_import.PDFStructuredTextExtractorWrapper;
import de.unibi.techfak.scie.util.CachedJCasUtil;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import junit.framework.TestCase;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class StructureAnnotatorTest extends TestCase {

	public StructureAnnotatorTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of process method, of class StructureAnnotator.
	 */
	public void testProcess() throws Exception {
		//first import a PDF document.
		final InputStream input = new FileInputStream("testdata/structure_test_document.pdf");
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final DocumentImporter instance = new PDFStructuredTextExtractorWrapper(input, jcas);
		//this importer automatically calls the StructureAnnotator.
		instance.doImport();

		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);
		//we expect only one page.
		final Collection<DocumentPage> pages = JCasUtil.select(jcas, DocumentPage.class);
		assertEquals(1, pages.size());

		final Iterator<TextBlock> blocks = JCasUtil.select(jcas, TextBlock.class).iterator();
		assertTrue(blocks.hasNext());
		//the first text block should be the introduction header.
		final String expectedHeader = "1 Introduction ";
		TextBlock block = blocks.next();
		assertStringEquals("\t" + expectedHeader, block.getCoveredText());
		assertTrue(block.getRelativeFontSize() > 1);
		//we expect a contained paragraph with the same content.
		List<Paragraph> paragraphs = jcasUtil.selectCovered(Paragraph.class,
				block.getBegin(), block.getEnd());
		assertEquals(1, paragraphs.size());
		assertStringEquals(expectedHeader, paragraphs.get(0).getCoveredText());

		//the second text block should be our main test text.
		final String[] expectedParagraphs = {
			"The oldest classical Greek and Latin writing had little or no spaces between words," +
			" and could be written in boustrophedon (alternating directions). Over time, text" +
			" direction (left to right) became standardized, and word dividers and terminal" +
			" punctuation became common. The first way to divide sentences into groups was the" +
			" original paragraphos, similar to an underscore at the beginning of the new group." +
			" The Greek paragraphos evolved into the pilcrow , which in English manuscripts in" +
			" the Middle Ages can be seen inserted inline between sentences. The hedera leaf has" +
			" also been used in the same way.",
			"In ancient manuscripts, another means to divide sentences in into paragraphs was a line" +
			" break (newline) followed by an initial at the beginning of the next paragraph. An" +
			" initial is an oversize capital letter, sometimes outdented beyond the margin of" +
			" text. This style can be seen, for example, in the original Old English manuscript" +
			" of Beowulf. Outdenting is still used in English typography, though not commonly." +
			"[4] Initials and rubrication were used in the Gutenberg Bible, the first major book" +
			" to be printed rather than hand-written.",
			"Modern English typography usually indicates a new paragraph by indenting the first line." +
			" This style can be seen in the (handwritten) United States Constitution from 1787." +
			" For additional ornamentation, a hedera leaf or other symbol can be added to the" +
			" inter-paragraph whitespace, or put in the indentation space.",
			"A second common modern English style is to use no indenting, but add vertical" +
			" whitespace to create ”block paragraphs”. On a typewriter, a double carriage return" +
			" produces a blank line for this purpose; professional typesetters may put in an" +
			" ar-bitrary vertical space by adjusting leading. This style is very common in" +
			" electronic formats, such as on the World Wide Web and email. "
		};
		assertTrue(blocks.hasNext());
		block = blocks.next();
		assertEquals(1.0, block.getRelativeFontSize(), 0);
		paragraphs = jcasUtil.selectCovered(Paragraph.class, block.getBegin(), block.getEnd());
		for (int i = 0; i < expectedParagraphs.length; i++) {
			assertStringEquals(expectedParagraphs[i], paragraphs.get(i).getCoveredText());
		}

		//the next block should be the methods and materials header.
		assertTrue(blocks.hasNext());
		block = blocks.next();
		final String matHeader = "2 Material and Methods ";
		assertStringEquals("\t" + matHeader, block.getCoveredText());
		paragraphs = jcasUtil.selectCovered(Paragraph.class, block.getBegin(), block.getEnd());
		assertEquals(1, paragraphs.size());
		assertStringEquals(matHeader, paragraphs.get(0).getCoveredText());

		//the next block should be the methods and materials content.
		assertTrue(blocks.hasNext());
		block = blocks.next();
		final String matText = "We used CO 2 injection, which is stupid 2 . ";
		assertStringEquals("\t" + matText, block.getCoveredText());
		paragraphs = jcasUtil.selectCovered(Paragraph.class, block.getBegin(), block.getEnd());
		assertEquals(1, paragraphs.size());
		assertStringEquals(matText, paragraphs.get(0).getCoveredText());
		//also we expect one sub- and one superscript.
		List<Subscript> subscripts =
				jcasUtil.selectCovered(Subscript.class, block.getBegin(), block.getEnd());
		//which should be the only one in the text.
		Collection<Subscript> expSubScripts = JCasUtil.select(jcas, Subscript.class);
		assertEquals(1, subscripts.size());
		assertEquals(1, expSubScripts.size());
		assertSame(expSubScripts.iterator().next(), subscripts.get(0));
		assertEquals("2", subscripts.get(0).getCoveredText());
		List<Superscript> superscripts =
				jcasUtil.selectCovered(Superscript.class, block.getBegin(), block.getEnd());
		//which should be the only one in the text.
		Collection<Superscript> expSuperScripts = JCasUtil.select(jcas, Superscript.class);
		assertEquals(1, superscripts.size());
		assertEquals(1, expSuperScripts.size());
		assertSame(expSuperScripts.iterator().next(), superscripts.get(0));
		assertEquals("2", superscripts.get(0).getCoveredText());

		//we do not check the rest of the text, as it has no new cases.
	}

	private static void assertStringEquals(final String exp, final String act) {
		assertEquals("Expected: \"" + exp + "\", but was: \"" + act + "\"", exp, act);
	}
}
