/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie.document_import;

import de.unibi.techfak.scie.document_import.DocumentImporter;
import de.unibi.techfak.scie.document_import.PDFStructuredTextExtractorWrapper;
import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.scie.annotators.abbreviation.AbbreviationAnnotator;
import de.unibi.techfak.scie.annotators.injury.InjuryAnnotator;
import de.unibi.techfak.scie.annotators.injury.VertebralPositionAnnotator;
import de.unibi.techfak.scie.annotators.injury.VertebralPositionRangeAnnotator;
import de.unibi.techfak.scie.annotators.ontology.OntologyAnnotator;
import de.unibi.techfak.scie.annotators.structure.HeaderAnnotator;
import de.unibi.techfak.scie.annotators.quantities.NumberAnnotator;
import de.unibi.techfak.scie.annotators.quantities.RangeAnnotator;
import de.unibi.techfak.scie.annotators.quantities.DurationAnnotator;
import de.unibi.techfak.scie.annotators.quantities.InexactQuantityAnnotator;
import de.unibi.techfak.scie.annotators.quantities.TimeAnnotator;
import de.unibi.techfak.scie.annotators.quantities.WordNumberAnnotator;
import de.unibi.techfak.scie.annotators.result.JudgementAnnotator;
import de.unibi.techfak.scie.annotators.result.ResultAnnotator;
import de.unibi.techfak.scie.annotators.result.PValueAnnotator;
import de.unibi.techfak.scie.annotators.structure.SectionAnnotator;
import de.unibi.techfak.scie.annotators.structure.StanfordTokenizer;
import de.unibi.techfak.scie.annotators.animal.ExactAgeAnnotator;
import de.unibi.techfak.scie.annotators.animal.LabGroupAnnotator;
import de.unibi.techfak.scie.annotators.animal.WeightAnnotator;
import de.unibi.techfak.scie.annotators.treatment.AntiBodyAnnotator;
import de.unibi.techfak.scie.annotators.treatment.DoseAnnotator;
import de.unibi.techfak.scie.annotators.treatment.TreatmentAnnotator;
import de.unibi.techfak.scie.db.DatabaseSingleton;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms;
import de.unibi.techfak.scie.analysis.AnalysisAnnotator;

import java.io.File;
import java.io.FileInputStream;

import junit.framework.TestCase;

import org.apache.uima.analysis_engine.AnalysisEngineDescription;
import org.apache.uima.fit.factory.AnalysisEngineFactory;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.pipeline.SimplePipeline;
import org.apache.uima.fit.util.CasIOUtil;
import org.apache.uima.jcas.JCas;

/**
 *
 * @author astoecke
 */
public class PDFDocumentImporterTest extends TestCase {

	public PDFDocumentImporterTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	private void runPipeline(File file) throws Exception {
		// Open the test document...
		FileInputStream input = new FileInputStream(file);

		// ...import it...
		JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final DocumentImporter instance = new PDFStructuredTextExtractorWrapper(input, jcas);

		instance.doImport();

		//set up a whole pipeline
		//structure
		final AnalysisEngineDescription stanfordTokenizer = AnalysisEngineFactory.
				createEngineDescription(StanfordTokenizer.class);
		final AnalysisEngineDescription headerAnnotator = AnalysisEngineFactory.
				createEngineDescription(HeaderAnnotator.class);
		final AnalysisEngineDescription sectionAnnotator = AnalysisEngineFactory.
				createEngineDescription(SectionAnnotator.class);

		//ontology
		final AnalysisEngineDescription ontologyAnnotator = AnalysisEngineFactory.
				createEngineDescription(OntologyAnnotator.class);

		//Abbreviation
		final AnalysisEngineDescription abbreviationAnnotator = AnalysisEngineFactory.
				createEngineDescription(AbbreviationAnnotator.class);

		//quantities
		final AnalysisEngineDescription numberAnnotator = AnalysisEngineFactory.
				createEngineDescription(NumberAnnotator.class);
		final AnalysisEngineDescription wordNumberAnnotator = AnalysisEngineFactory.
				createEngineDescription(WordNumberAnnotator.class);
		final AnalysisEngineDescription inexactQuantityAnnotator = AnalysisEngineFactory.
				createEngineDescription(InexactQuantityAnnotator.class);
		final AnalysisEngineDescription rangeAnnotator = AnalysisEngineFactory.
				createEngineDescription(RangeAnnotator.class);
		final AnalysisEngineDescription timeAnnotator = AnalysisEngineFactory.
				createEngineDescription(TimeAnnotator.class);
		final AnalysisEngineDescription durationAnnotator = AnalysisEngineFactory.
				createEngineDescription(DurationAnnotator.class);

		//animals
		final AnalysisEngineDescription exactAgeAnnotator = AnalysisEngineFactory.
				createEngineDescription(ExactAgeAnnotator.class);
		final AnalysisEngineDescription weightAnnotator = AnalysisEngineFactory.
				createEngineDescription(WeightAnnotator.class);
		final AnalysisEngineDescription labGroupAnnotator = AnalysisEngineFactory.
				createEngineDescription(LabGroupAnnotator.class);

		//treatments
		final AnalysisEngineDescription antibodyAnnotator = AnalysisEngineFactory.
				createEngineDescription(AntiBodyAnnotator.class);
		final AnalysisEngineDescription doseAnnotator = AnalysisEngineFactory.
				createEngineDescription(DoseAnnotator.class);
		final AnalysisEngineDescription treatmentAnnotator = AnalysisEngineFactory.
				createEngineDescription(TreatmentAnnotator.class);

		//injuries
		final AnalysisEngineDescription vpAnnotator = AnalysisEngineFactory.
				createEngineDescription(VertebralPositionAnnotator.class);
		final AnalysisEngineDescription vprAnnotator = AnalysisEngineFactory.
				createEngineDescription(VertebralPositionRangeAnnotator.class);
		final AnalysisEngineDescription injuryAnnotator = AnalysisEngineFactory.
				createEngineDescription(InjuryAnnotator.class);
		//results
		final AnalysisEngineDescription judgementAnnotator = AnalysisEngineFactory.
				createEngineDescription(JudgementAnnotator.class);
		final AnalysisEngineDescription pvalueAnnotator = AnalysisEngineFactory.
				createEngineDescription(PValueAnnotator.class);
		final AnalysisEngineDescription resultAnnotator = AnalysisEngineFactory.
				createEngineDescription(ResultAnnotator.class);

		final AnalysisEngineDescription AnalysisAnotator = AnalysisEngineFactory.
				createEngineDescription(AnalysisAnnotator.class);

		//reset counter for annotationId
		AnnotationCounter.reset();

		//and run it
		SimplePipeline.runPipeline(jcas,
				//ontology
				ontologyAnnotator,
				//structure
				stanfordTokenizer,
				//quantities
				numberAnnotator, wordNumberAnnotator, rangeAnnotator,
				timeAnnotator, durationAnnotator,
				//Abbreviation
				abbreviationAnnotator,
				//inexact qunatifiers
				inexactQuantityAnnotator,
				//lab animals
				exactAgeAnnotator, weightAnnotator, labGroupAnnotator,
				//structure again, because the ontology annotator has to process before headers
				//can be recognized
				headerAnnotator, sectionAnnotator,
				//injuries
				vpAnnotator, vprAnnotator, injuryAnnotator,
				//treatments
				doseAnnotator, antibodyAnnotator, treatmentAnnotator,
				//results
				judgementAnnotator, pvalueAnnotator, resultAnnotator,
				//analysis
				AnalysisAnotator);

		DictionaryAlgorithms.purgeWordsCache();

		// ...and dump the result
		String inFile = file.getName();
		File outFile = new File(
				"testdata/XCAS/" + inFile.substring(0, inFile.lastIndexOf('.')) + ".xml");
		(new File("testdata/XCAS")).mkdirs();

		CasIOUtil.writeXCas(jcas, outFile);
	}

	/**
	 * Test of doImport method, of class PDFDocumentImporter.
	 */
	public void testDoImport() throws Exception {
		System.out.println("doImport");

		DatabaseSingleton.initialize("./config.xml");

		// Iterate over all ".pdf" files in the testdata folder and sub folders
		File folder = new File("./testdata");
		File[] files = folder.listFiles();
		for (File file : files) {
			if (file.getName().endsWith(".pdf")) {
				System.out.println("Processing file " + file.getName());
				runPipeline(file);
			}
			if (file.isDirectory()) {
				File[] files2 = file.listFiles();
				for (File file2 : files2) {
					if (file2.getName().endsWith(".pdf")) {
						System.out.println("Processing file " + file2.getName());
						runPipeline(file2);
					}
				}
			}
		}
		assertEquals("This test should not fail.", true, true);
	}
}
