/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.DictionaryAlgorithms;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms.WordPosition;
import java.util.Set;
import junit.framework.TestCase;

/**
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public class DictionaryAlgorithmsTest extends TestCase {

	public DictionaryAlgorithmsTest(String testName) {
		super(testName);
	}

	/**
	 * Test of isStopWord method, of class DictionaryAlgorithms.
	 */
	public void testIsStopWord() {
		System.out.println("isStopWord");
		assertTrue("the", DictionaryAlgorithms.isStopWord("the"));
		assertTrue("then", DictionaryAlgorithms.isStopWord("then"));
		assertTrue("this", DictionaryAlgorithms.isStopWord("this"));
		assertTrue("thus", DictionaryAlgorithms.isStopWord("thus"));
		assertTrue("a", DictionaryAlgorithms.isStopWord("a"));
		assertTrue("an", DictionaryAlgorithms.isStopWord("an"));
		assertTrue("is", DictionaryAlgorithms.isStopWord("is"));
		assertTrue("we", DictionaryAlgorithms.isStopWord("we"));
		assertTrue("so", DictionaryAlgorithms.isStopWord("so"));
		assertTrue("cat", !DictionaryAlgorithms.isStopWord("cat"));
		assertTrue("teddy", !DictionaryAlgorithms.isStopWord("teddy"));
	}

	/**
	 * Test of wordStem method, of class DictionaryAlgorithms.
	 */
	public void testWordStem() {
		System.out.println("wordStem");
		assertEquals("banking", "bank", DictionaryAlgorithms
			.wordStem("banking"));
		assertEquals("banked", "bank", DictionaryAlgorithms
			.wordStem("banked"));
		assertEquals("banks", "bank", DictionaryAlgorithms
			.wordStem("banks"));
		assertEquals("trosophilae", "trosophil", DictionaryAlgorithms
			.wordStem("trosophilae"));
		assertEquals("greenish", "green", DictionaryAlgorithms
			.wordStem("greenish"));
		assertEquals("adequatly", "adequat", DictionaryAlgorithms
			.wordStem("adequatly"));
		assertEquals("smaller", "small", DictionaryAlgorithms
			.wordStem("smaller"));
		assertEquals("enteral", "ent", DictionaryAlgorithms
			.wordStem("enteral"));
		assertEquals("intracavernos", "intracavern", DictionaryAlgorithms
			.wordStem("intracavernos"));
		assertEquals("intracavernous", "intracavern", DictionaryAlgorithms
			.wordStem("intracavernous"));
		assertEquals("increasing", DictionaryAlgorithms.wordStem("increase"),
			DictionaryAlgorithms.wordStem("increasing"));

	}

	/**
	 * Test of editDistance method, of class DictionaryAlgorithms.
	 */
	public void testEditDistance() {
		System.out.println("editDistance");
		assertEquals("gabor <-> gubo", 2, DictionaryAlgorithms
			.editDistance("gabor", "gubo"));
		assertEquals("gubo <-> gabor", 2, DictionaryAlgorithms
			.editDistance("gubo", "gabor"));
		assertEquals("test <-> fest", 1, DictionaryAlgorithms
			.editDistance("test", "fest"));
		assertEquals("rest <-> rost", 1, DictionaryAlgorithms
			.editDistance("rest", "rost"));
		assertEquals("rest <-> rest", 0, DictionaryAlgorithms
			.editDistance("rest", "rest"));
	}

	/**
	 * Test of matchProbability method, of class DictionaryAlgorithms.
	 */
	public void testMatchProbability() {
		System.out.println("matchProbability");
		assertEquals("gabor <-> gubo", 0.6f, DictionaryAlgorithms
			.matchProbability("gabor", "gubo"));
		assertEquals("gubo <-> gabor", 0.6f, DictionaryAlgorithms
			.matchProbability("gubo", "gabor"));
		assertEquals("test <-> fest", 0.75f, DictionaryAlgorithms
			.matchProbability("test", "fest"));
		assertEquals("rest <-> rost", 0.75f, DictionaryAlgorithms
			.matchProbability("rest", "rost"));
		assertEquals("rest <-> rest", 1.0f, DictionaryAlgorithms
			.matchProbability("rest", "rest"));
	}

	/**
	 * Test of normalizeWord method, of class DictionaryAlgorithms.
	 */
	public void testCalculateWordKeys() {
		System.out.println("calculateWordKeys");
		assertEquals("BAnking", "bank", DictionaryAlgorithms
			.calculateWordKeys("BAnking")[0]);
		assertEquals("   b@@ANKED  ", "bank", DictionaryAlgorithms
			.calculateWordKeys("   b@@ANKED  ")[0]);
		assertEquals("Bank's", "bank", DictionaryAlgorithms
			.calculateWordKeys("Bank's")[0]);
		assertEquals("trosophilae", "trosophil", DictionaryAlgorithms
			.calculateWordKeys("trosophilae")[0]);
		assertEquals("greenish", "green", DictionaryAlgorithms
			.calculateWordKeys("greenish")[0]);
		assertEquals(" !!!this.;-..--", "this", DictionaryAlgorithms
			.calculateWordKeys(" !!!this.;-..--\n\n\t\t")[0]);
		assertEquals("we", "we", DictionaryAlgorithms
			.calculateWordKeys("we")[0]);
		assertEquals("weish", "weish", DictionaryAlgorithms
			.calculateWordKeys("weish")[0]);
		assertEquals("@123!ÜÖÄ", "@123!üöä", DictionaryAlgorithms
			.calculateWordKeys("@123!ÜÖÄ")[0]);
		assertEquals("passenger", "passeng", DictionaryAlgorithms
			.calculateWordKeys("passengers")[0]);
	}

	/**
	 * Test of pseudoPhoneticKeys method, of class DictionaryAlgorithms.
	 */
	public void testPseudoPhoneticKeys() {
		System.out.println("pseudoPhoneticKeys");
		assertEquals("empty string", 0,
			DictionaryAlgorithms.calculateWordKeys("").length);
		assertEquals("null string", 0,
			DictionaryAlgorithms.calculateWordKeys(null).length);
	}

	/**
	 * Tests the "getWords" function.
	 */
	public void testGetWords() {
		System.out.println("getWords");

		final String s1 = "ground-breaking news, moon stolen!";
		Set<WordPosition> words = DictionaryAlgorithms.getWords(s1, false);
		assertEquals("Word count", 8, words.size());
		assertTrue("ground", words.contains(new WordPosition(s1, 0, 6)));
		assertTrue("breaking", words.contains(new WordPosition(s1, 7, 15)));
		assertTrue("ground-breaking", words.contains(new WordPosition(s1, 0, 15)));
		assertTrue("news", words.contains(new WordPosition(s1, 16, 20)));
		assertTrue("news,", words.contains(new WordPosition(s1, 16, 21)));
		assertTrue("moon", words.contains(new WordPosition(s1, 22, 26)));
		assertTrue("stolen", words.contains(new WordPosition(s1, 27, 33)));
		assertTrue("stolen!", words.contains(new WordPosition(s1, 27, 34)));

		//test caching for the same object
		assertSame(words, DictionaryAlgorithms.getWords(s1, false));
	}

}
