/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.dictionary_name_finder;

import de.unibi.techfak.scie.ontology.dictionary_name_finder.DictionaryNameFinder;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms.WordPosition;
import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.driver.implementation.MemoryDatabaseDriver;
import de.unibi.techfak.scie.probability_model.CharBasedSyntacticProbabilityModel;
import java.util.Collection;
import junit.framework.TestCase;

/**
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public class DictionaryNameFinderTest extends TestCase {

	public DictionaryNameFinderTest(String testName) {
		super(testName);
	}

	private static final String test1 = "Mr. J. Frenzy confirmed: There was more than "
			+ "a handful of apes on flight RE-385:Baltimore.";
	private static final String test2 = "Mr.F. furthermore revealed, that the apes were not "
			+ "only held as passengers, but also the ones who actually flew "
			+ "the airplane.";

	private static final double SIGMA = 50;

	private Collection<DictionaryMatch> dictionaryNameFinderTestUtilMatches(
			String text, NodeName name) {
		// Create a memory database and add a node with the given node name
		MemoryDatabaseDriver driver = new MemoryDatabaseDriver();
		OntologyDatabase db = new OntologyDatabase(driver);
		OntologyHandle onto = db.getOntology("test");
		db.storeNode(new Node(db, onto, 0, name));

		// Try to find the name within the given text
		DictionaryNameFinder finder = new DictionaryNameFinder(db, text, 0.4,
				new CharBasedSyntacticProbabilityModel(SIGMA));
		finder.processDocumentText();
		return finder.calculateMatches();
	}

	private void dictionaryNameFinderTestUtil(String text, NodeName name, WordPosition[] positions) {		
		Collection<DictionaryMatch> matches = dictionaryNameFinderTestUtilMatches(text, name);
		int matchCount = 0;
		String mode = name.getComparisonMode().name();
		for (DictionaryMatch match : matches) {
			final int b = match.getBegin();
			final int e = match.getEnd();
			final String t = text.substring(b, e);
			System.out.println("Found " + t + " @ " + mode + " (" + b +  ", " + e + ")");
			boolean found = false;
			for (int i = 0; i < positions.length; i++) {
				if (positions[i] != null && positions[i].getBegin() == b && positions[i].getEnd() == e) {
					positions[i] = null;
					found = true;
					matchCount++;
					break;
				}
			}
			assertTrue("Match not found in position list (" +  t + ") @ "
					+ mode + ")", found);
		}
		assertEquals("Not all matches were found! (" + name.getName() + " @ "
				+ mode + ")", positions.length, matchCount);
	}

	public void testComparisonModesNormalization() {
		final String test = "ontolog ontology ontolog31y ontolog?31-y ONtOlog?31-Y";
		dictionaryNameFinderTestUtil(
				test, new NodeName("ONtOlog?31-Y", ComparisonMode.FUZZY),
				new WordPosition[]{
					new WordPosition(0, 7),
					new WordPosition(8, 16),
					new WordPosition(17, 27),
					new WordPosition(28, 35),
					new WordPosition(28, 40),
					new WordPosition(41, 48),
					new WordPosition(41, 53),});
		dictionaryNameFinderTestUtil(
				test, new NodeName("ONtOlog?31-Y", ComparisonMode.SOFT),
				new WordPosition[]{
					new WordPosition(8, 16),
					new WordPosition(17, 27),
					new WordPosition(28, 40),
					new WordPosition(41, 53),});
		dictionaryNameFinderTestUtil(
				test, new NodeName("ONtOlog?31-Y", ComparisonMode.CLEAN),
				new WordPosition[]{
					new WordPosition(17, 27),
					new WordPosition(28, 40),
					new WordPosition(41, 53),});
		dictionaryNameFinderTestUtil(
				test, new NodeName("ONtOlog?31-Y", ComparisonMode.LOWERCASE),
				new WordPosition[]{
					new WordPosition(28, 40),
					new WordPosition(41, 53),});
		dictionaryNameFinderTestUtil(
				test, new NodeName("ONtOlog?31-Y", ComparisonMode.STRICT),
				new WordPosition[]{
					new WordPosition(41, 53),});
	}

	public void testComparisonModesSingleWord() {
		// "confirm" must not be found in any other comparison mode
		for (ComparisonMode mode : ComparisonMode.values()) {
			if (mode == ComparisonMode.FUZZY) {
				dictionaryNameFinderTestUtil(
						test1, new NodeName("confirm", ComparisonMode.FUZZY),
						new WordPosition[]{new WordPosition(14, 23)});
			} else {
				dictionaryNameFinderTestUtil(
						test1, new NodeName("confirm", mode),
						new WordPosition[]{});
			}
		}

		// However, if we search for "confirmed" the word should be found in all
		// comparison modes -- except for "NONE"
		for (ComparisonMode mode : ComparisonMode.values()) {
			if (mode == ComparisonMode.NONE) {
				dictionaryNameFinderTestUtil(
						test1, new NodeName("confirmed", mode),
						new WordPosition[]{});
			} else {
				dictionaryNameFinderTestUtil(
						test1, new NodeName("confirmed", mode),
						new WordPosition[]{new WordPosition(14, 23)});
			}
		}

		// Fuzzy match for "passenger" should work -- this one is tricky and
		// lead to problems as "passengers" is reduced to "passenger" when the
		// text is parsed, but "passenger" is reduced to "passeng". Now the word
		// stem is calculated recursively.
		dictionaryNameFinderTestUtil(
				test2, new NodeName("passenger", ComparisonMode.FUZZY),
				new WordPosition[]{new WordPosition(64, 74)});
	}

	public void testComparisonModesMultipleWords() {
		// Search for "Mr. F." should work in all modes up to and including "CLEAN"
		// then it should no longer be found. In the "FUZZY" mode this particular
		// example is not found, as the single words are too short.
		for (ComparisonMode mode : ComparisonMode.values()) {
			if (mode.getValue() >= ComparisonMode.SOFT.getValue()
					&& mode.getValue() <= ComparisonMode.CLEAN.getValue()) {
				dictionaryNameFinderTestUtil(
						test2, new NodeName("Mr. F.", mode),
						new WordPosition[]{new WordPosition(0, 4)});
			} else {
				dictionaryNameFinderTestUtil(
						test2, new NodeName("Mr. F.", mode),
						new WordPosition[]{});
			}
		}

		// The text "actually flew the airplaine" should be found in all modes
		for (ComparisonMode mode : ComparisonMode.values()) {
			if (mode == ComparisonMode.NONE) {
				dictionaryNameFinderTestUtil(
						test2, new NodeName("actually flew the airplane", mode),
						new WordPosition[]{});
			} else {
				dictionaryNameFinderTestUtil(
						test2, new NodeName("actually flew the airplane", mode),
						new WordPosition[]{new WordPosition(98, 124)});
			}
		}

		// Search for "Mr. Frenzy" in FUZZY mode should match "Frenzy" in "Mr. J. Frenzy"
		dictionaryNameFinderTestUtil(
				test1, new NodeName("Mr. Frenzy", ComparisonMode.FUZZY),
				new WordPosition[]{new WordPosition(7, 13)});
	}

	public void testAmbiguousWordEndings() {	
		dictionaryNameFinderTestUtil(
				test1, new NodeName("RE-385:Baltimore", ComparisonMode.STRICT),
				new WordPosition[]{new WordPosition(73, 89)});
		dictionaryNameFinderTestUtil(
				test1, new NodeName("RE-385:Baltimore.", ComparisonMode.STRICT),
				new WordPosition[]{new WordPosition(73, 90)});

		final String test = "The rats were anesthetized with ketamine (50 mg/kg) and rompun (2 mg/kg, i.p.).";
		dictionaryNameFinderTestUtil(
				test, new NodeName("i.p.", ComparisonMode.LOWERCASE),
				new WordPosition[]{new WordPosition(73, 77)});
		dictionaryNameFinderTestUtil(
				test, new NodeName("mg/kg", ComparisonMode.LOWERCASE),
				new WordPosition[]{
					new WordPosition(45, 50),
					new WordPosition(66, 71),});
	}

	public void testFuzzyProbabilities() {
		final String test = "Increased growth factor expression and cell proliferation";
	
		Collection<DictionaryMatch> matches = dictionaryNameFinderTestUtilMatches(
				test, new NodeName("cell growth factor"));
	}

	public void testMultiWordMatching() {
		final String test = "kleine kleine kleine Maus";

		dictionaryNameFinderTestUtil(
				test, new NodeName("kleine kleine", ComparisonMode.FUZZY),
				new WordPosition[]{
					new WordPosition(0, 6),
					new WordPosition(7, 13),
					new WordPosition(14, 20),
					new WordPosition(0, 13),
					new WordPosition(7, 20),
					new WordPosition(0, 20)});

		dictionaryNameFinderTestUtil(
				test, new NodeName("kleine kleine", ComparisonMode.STRICT),
				new WordPosition[]{
					new WordPosition(0, 13),
					new WordPosition(7, 20),});
	}

	/**
	 * Test of calculateMatches method, of class DictionaryNameFinder.
	 *
	 * Note: This test is only for debuging purposes and not an actual
	 * regression test.
	 */
	public void testCalculateMatches() {
		final String testString = "Mr. J. Frenzy confirmed that there was more than "
				+ "a handful of apes on flight RE-385:Baltimore. Mr.F. furthermore "
				+ "revealed, that the apes were not only held as passengers, but "
				+ "also the ones who actually flew the airplane. J.F. likes to "
				+ "stress, that the monkeys did a very good job and --at no time-- "
				+ "did he feel as if his life was in danger; Especially the 3-headed "
				+ "monkey kept control over the machinery as if he was in a "
				+ "sensual frenzy. Mr. frenzy also added --and please don't listen to this "
				+ "nonsense, we've only included it in this report for the sake "
				+ "of completeness-- that in the end it was the Airforce special "
				+ "force unit US11 which started to cause all the chaos which "
				+ "finally lead to the destruction of RE-385 and caused the death "
				+ "of 381 people, twelve monkeys and one Bengal Tiger.";

		// Fill the ontology database with some names to search for
		MemoryDatabaseDriver driver = new MemoryDatabaseDriver();
		OntologyDatabase db = new OntologyDatabase(driver);
		OntologyHandle onto = db.getOntology("re385");
		NodeName names[] = new NodeName[]{
			new NodeName("passenger", ComparisonMode.FUZZY),
			new NodeName("people", ComparisonMode.FUZZY),
			new NodeName("Mr. Frenzy", ComparisonMode.FUZZY),
			new NodeName("nonsense", ComparisonMode.SOFT),
			new NodeName("Mr.", ComparisonMode.SOFT),
			new NodeName("Mr. Frenzy", ComparisonMode.SOFT),
			new NodeName("no time", ComparisonMode.SOFT),
			new NodeName("bengal tiger", ComparisonMode.CLEAN),
			new NodeName("3-headed", ComparisonMode.CLEAN),
			new NodeName("Frenzy", ComparisonMode.STRICT),
			new NodeName("J.", ComparisonMode.STRICT),
			new NodeName("RE", ComparisonMode.STRICT),
			new NodeName("RE-385", ComparisonMode.STRICT),
			new NodeName("RE-385:Baltimore", ComparisonMode.STRICT),};
		int i = 0;
		for (NodeName name : names) {
			db.storeNode(new NodeDescriptor(onto, i++, name));
		}

		DictionaryNameFinder finder = new DictionaryNameFinder(db, testString, 0.25,
				new CharBasedSyntacticProbabilityModel(SIGMA));
		finder.processDocumentText();
		Collection<DictionaryMatch> matches = finder.calculateMatches();
		for (DictionaryMatch match : matches) {
			final int b = match.getBegin();
			final int e = match.getEnd();
			final String s = testString.substring(b, e);
			final NodeName name = match.getNameDescr();
			System.out.println(s + " <-> " + name.getName() + " @ " + name.getComparisonMode().toString());
		}
	}

}
