/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.probability_model;

import de.unibi.techfak.scie.probability_model.CharBasedSyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.AnnotationCompositionUtilTest.TrivialAnnotationCandidate;
import junit.framework.TestCase;

/**
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class CharBasedSyntacticProbabilityModelTest extends TestCase {

	public CharBasedSyntacticProbabilityModelTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateNormalDistribution method, of class
	 * CharBasedSyntacticProbabilityModel.
	 */
	public void testCalculateNormalDistribution() {
		final int sigma = 10;
		final CharBasedSyntacticProbabilityModel instance =
				new CharBasedSyntacticProbabilityModel(sigma);
		assertEquals(1, instance.calculateNormalDistribution(0), 0);
		double lastVal = 1;
		for (int dist = 1; dist < instance.getMaximumCharBasedRange(); dist++) {
			double current = instance.calculateNormalDistribution(dist);
			assertTrue(current < lastVal);
			assertTrue(current > 0);
			lastVal = current;
		}
		assertEquals(0, lastVal, 0.05);
	}

	/**
	 * Test of calculateSyntacticProbability method, of class
	 * CharBasedSyntacticProbabilityModel.
	 */
	public void testCalculateSyntacticProbability() {
		//create 
		final int maxDistance = 10;
		final double sigma = 20;
		final CharBasedSyntacticProbabilityModel instance =
				new CharBasedSyntacticProbabilityModel(sigma);
		//initialize core.
		final TrivialAnnotationCandidate a = new TrivialAnnotationCandidate();
		a.setBegin(maxDistance);
		a.setEnd(maxDistance + 1);
		/*
		 * We want to test the following things:
		 * a.) the syntatic probability should get lower if we increase the distance.
		 * b.) it should never be above 1.
		 * c.) it should be equal to the value of the normal distribution with the correct sigma and
		 * the input distance.
		 * d.) it should be symmetric for negative and positive distances and the order of the input
		 * operands should not matter either.
		 */

		double lastProb = 1;
		for (int d = 1; d <= maxDistance; d++) {
			//create one annotation before
			final TrivialAnnotationCandidate b =
					new TrivialAnnotationCandidate();
			b.setBegin(a.getEnd() + d);
			b.setEnd(a.getEnd() + d + 1);
			//... one annotation afterwards
			final TrivialAnnotationCandidate c =
					new TrivialAnnotationCandidate();
			c.setBegin(a.getBegin() - d - 1);
			c.setEnd(a.getBegin() - d);
			//... one covering annotation with shorter distance of the starts
			final TrivialAnnotationCandidate coveringA =
					new TrivialAnnotationCandidate();
			coveringA.setBegin(a.getBegin() - d);
			coveringA.setEnd(a.getEnd() + 3 * d);
			//... and one covering annotation with shorter distance of the ends
			final TrivialAnnotationCandidate coveringB =
					new TrivialAnnotationCandidate();
			coveringB.setBegin(a.getBegin() - 3 * d);
			coveringB.setEnd(a.getEnd() + d);

			//check results
			final double currentProb = instance.calculateSyntacticProbability(a,
					b);
			assertTrue(currentProb < lastProb);
			assertEquals(currentProb, instance.
					calculateNormalDistribution(d), 0);
			assertEquals(currentProb, instance.
					calculateSyntacticProbability(b, a));
			assertEquals(currentProb, instance.
					calculateSyntacticProbability(a, c));
			assertEquals(currentProb, instance.
					calculateSyntacticProbability(a, coveringA));
			assertEquals(currentProb, instance.
					calculateSyntacticProbability(a, coveringB));
			lastProb = currentProb;
		}
	}

	/**
	 * Test of getMaximumCharBasedRange method, of class
	 * CharBasedSyntacticProbabilityModel.
	 */
	public void testGetMaximumCharBasedRange() {
		final int sigma = 30;

		final CharBasedSyntacticProbabilityModel instance =
				new CharBasedSyntacticProbabilityModel(sigma);
		int expResult = 90;
		int result = instance.getMaximumCharBasedRange();
		assertEquals(expResult, result);
	}
}
