/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.util;

import de.unibi.techfak.scie.util.CachedJCasUtil;
import de.unibi.techfak.scie.annotators.structure.StanfordTokenizer;
import de.unibi.techfak.isy.lppba.descriptors.Token;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import junit.framework.TestCase;
import org.apache.uima.UIMAException;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 *
 * @author socnow
 */
public class CachedJCasUtilTest extends TestCase {

	public CachedJCasUtilTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of getInstance method, of class BufferedJCasUtil.
	 */
	public void testGetInstance() throws UIMAException {
		final JCas jcas1 = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final JCas jcas2 = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		CachedJCasUtil instance11 = CachedJCasUtil.getInstance(jcas1);
		assertNotNull(instance11);
		CachedJCasUtil instance12 = CachedJCasUtil.getInstance(jcas1);
		assertSame(instance11, instance12);

		CachedJCasUtil instance2 = CachedJCasUtil.getInstance(jcas2);
		assertNotSame(instance11, instance2);
	}

	/**
	 * Test of select method, of class BufferedJCasUtil.
	 */
	public void testSelect() throws UIMAException {
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final String text = "This is a test text.";
		jcas.setDocumentText(text);

		//find all words.
		final StanfordTokenizer st = new StanfordTokenizer();
		st.process(jcas);

		//get the correct instance.
		final CachedJCasUtil instance = CachedJCasUtil.getInstance(jcas);
		//then check whether regular select and cached select have the same result.
		Iterator<Token> expected = JCasUtil.select(jcas, Token.class).
				iterator();
		Iterator<Token> actual = instance.select(Token.class).iterator();
		while (expected.hasNext()) {
			assertTrue(actual.hasNext());
			assertSame(expected.next(), actual.next());
		}
	}

	/**
	 * Test of selectCovered method, of class BufferedJCasUtil.
	 */
	public void testSelectCovered() throws UIMAException {
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final String text = "This is a test text.";
		jcas.setDocumentText(text);
		//create some arbitrary annotations that are all covered in the same regions.
		final Token anno1 = new Token(jcas, 2, 4);
		anno1.addToIndexes();
		final Token anno2 = new Token(jcas, 2, 5);
		anno2.addToIndexes();
		final Token anno3 = new Token(jcas, 3, 4);
		anno3.addToIndexes();
		final ArrayList<Token> expected = new ArrayList<>();
		expected.add(anno1);
		expected.add(anno2);
		expected.add(anno3);
		//also add trash annotations that we do not want to find.
		final Token anno4 = new Token(jcas, 1, 4);
		anno4.addToIndexes();
		final Token anno5 = new Token(jcas, 2, 9);
		anno5.addToIndexes();


		//get the instance.
		final CachedJCasUtil instance = CachedJCasUtil.getInstance(jcas);
		//check results.
		final List<Token> actual = instance.selectCovered(Token.class, 2, 5);
		assertEquals(expected.size(), actual.size());
		for (int i = 0; i < expected.size(); i++) {
			assertSame(expected.get(i), actual.get(i));
		}
	}

	/**
	 * Test of selectCovering method, of class BufferedJCasUtil.
	 */
	public void testSelectCovering() throws UIMAException {
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final String text = "This is a test text.";
		jcas.setDocumentText(text);
		//create some arbitrary annotations that all cover the same region.
		final Token anno1 = new Token(jcas, 1, 5);
		anno1.addToIndexes();
		final Token anno2 = new Token(jcas, 1, 8);
		anno2.addToIndexes();
		final Token anno3 = new Token(jcas, 3, 7);
		anno3.addToIndexes();
		final ArrayList<Token> expected = new ArrayList<>();
		expected.add(anno1);
		expected.add(anno2);
		expected.add(anno3);
		//also add trash annotations that we do not want to find.
		final Token anno4 = new Token(jcas, 3, 4);
		anno4.addToIndexes();
		final Token anno5 = new Token(jcas, 4, 5);
		anno5.addToIndexes();

		//get the instance.
		final CachedJCasUtil instance = CachedJCasUtil.getInstance(jcas);
		//check results.
		final List<Token> actual = instance.selectCovering(Token.class, 3, 5);
		assertEquals(expected.size(), actual.size());
		for (int i = 0; i < expected.size(); i++) {
			assertSame(expected.get(i), actual.get(i));
		}
	}
	
	/**
	 * Test of selectCovering method, of class BufferedJCasUtil.
	 */
	public void testSelectOverlapping() throws UIMAException {
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		final String text = "This is a test text.";
		jcas.setDocumentText(text);
		//create some arbitrary annotations that all cover the same region.
		final Token anno1 = new Token(jcas, 1, 7);
		anno1.addToIndexes();
		final Token anno2 = new Token(jcas, 3, 5);
		anno2.addToIndexes();
		final Token anno3 = new Token(jcas, 4, 8);
		anno3.addToIndexes();
		final ArrayList<Token> expected = new ArrayList<>();
		expected.add(anno1);
		expected.add(anno2);
		expected.add(anno3);
		//also add trash annotations that we do not want to find.
		final Token anno4 = new Token(jcas, 1, 3);
		anno4.addToIndexes();
		final Token anno5 = new Token(jcas, 6, 8);
		anno5.addToIndexes();

		//get the instance.
		final CachedJCasUtil instance = CachedJCasUtil.getInstance(jcas);
		//check results.
		final List<Token> actual = instance.selectCovering(Token.class, 4, 5);
		assertEquals(expected.size(), actual.size());
		for (int i = 0; i < expected.size(); i++) {
			assertSame(expected.get(i), actual.get(i));
		}
	}
}
